# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado
"""

# from os.path import splitext, basename
import logging
import uuid
import six
from wbia.web.routes_ajax import image_src
from wbia.control import controller_inject
import utool as ut
import wbia.constants as const

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

register_api = controller_inject.get_wbia_flask_api(__name__)


@register_api('/api/imageset/json/', methods=['POST'])
def add_imagesets_json(
    ibs,
    imageset_text_list,
    imageset_uuid_list=None,
    config_rowid_list=None,
    imageset_notes_list=None,
    imageset_occurence_flag_list=None,
):
    r"""
    Adds a list of imagesets.

    Args:
        imagesettext_list (list):
        imageset_uuid_list (list):
        config_rowid_list (list):
        notes_list (list):

    Returns:
        imageset_uuid_list (list): added imageset uuids

    RESTful:
        Method: POST
        URL:    /api/imageset/json/
    """
    imageset_rowid_list = ibs.add_imagesets(
        imageset_text_list,
        imageset_uuid_list=imageset_uuid_list,
        occurence_flag_list=imageset_occurence_flag_list,
        config_rowid_list=config_rowid_list,
        notes_list=imageset_notes_list,
    )
    imageset_uuid_list = ibs.get_imageset_uuid(imageset_rowid_list)
    return imageset_uuid_list


# @register_api('/api/image/json/', methods=['POST'])
# def add_images_json(ibs, image_uri_list, image_uuid_list, image_width_list,
#                     image_height_list, image_orig_name_list=None, image_ext_list=None,
#                     image_time_posix_list=None, image_gps_lat_list=None,
#                     image_gps_lon_list=None, image_orientation_list=None,
#                     image_notes_list=None, **kwargs):
#     """
#     REST:
#         Method: POST
#         URL: /api/image/json/

#     Ignore:
#         sudo pip install boto

#     Args:
#         image_uri_list (list) : list of string image uris, most likely HTTP(S) or S3
#             encoded URLs.  Alternatively, this can be a list of dictionaries (JSON
#             objects) that specify AWS S3 stored assets.  An example below:

#                 image_uri_list = [
#                     'http://domain.com/example/asset1.png',
#                     '/home/example/Desktop/example/asset2.jpg',
#                     's3://s3.amazon.com/example-bucket-2/asset1-in-bucket-2.tif',
#                     {
#                         'bucket'          : 'example-bucket-1',
#                         'key'             : 'example/asset1.png',
#                         'auth_domain'     : None,  # Uses 127.0.0.1
#                         'auth_access_id'  : None,  # Uses system default
#                         'auth_secret_key' : None,  # Uses system default
#                     },
#                     {
#                         'bucket' : 'example-bucket-1',
#                         'key'    : 'example/asset2.jpg',
#                         # if unspecified, auth uses 127.0.0.1 and system defaults
#                     },
#                     {
#                         'bucket'          : 'example-bucket-2',
#                         'key'             : 'example/asset1-in-bucket-2.tif',
#                         'auth_domain'     : 's3.amazon.com',
#                         'auth_access_id'  : '____________________',
#                         'auth_secret_key' : '________________________________________',
#                     },
#                 ]

#             Note that you cannot specify AWS authentication access ids or secret keys
#             using string uri's.  For specific authentication methods, please use the
#             latter list of dictionaries.

#         image_uuid_list (list of str) : list of image UUIDs to be used in IBEIS IA
#         image_width_list (list of int) : list of image widths
#         image_height_list (list of int) : list of image heights
#         image_orig_name_list (list of str): list of original image names
#         image_ext_list (list of str): list of original image names
#         image_time_posix_list (list of int): list of image's POSIX timestamps
#         image_gps_lat_list (list of float): list of image's GPS latitude values
#         image_gps_lon_list (list of float): list of image's GPS longitude values
#         image_orientation_list (list of int): list of image's orientation flags
#         image_notes_list (list of str) : optional list of any related notes with
#             the images
#         **kwargs : key-value pairs passed to the ibs.add_images() function.

#     CommandLine:
#         python -m wbia.web.apis_json --test-add_images_json


#         ,"bucket":"flukebook-prod-asset-store","key":""

#     Example:
#         >>> # WEB_DOCTEST
#         >>> from wbia.control.IBEISControl import *  # NOQA
#         >>> import wbia
#         >>> import uuid
#         >>> web_instance = wbia.opendb(db='testdb1')
#         >>> _payload = {
#         >>>     'image_uri_list': [
#         >>>         'https://upload.wikimedia.org/wikipedia/commons/4/49/Zebra_running_Ngorongoro.jpg',
#         >>>         {
#         >>>             'bucket'          : 'test-asset-store',
#         >>>             'key'             : 'caribwhale/20130903-JAC-0002.JPG',
#         >>>         },
#         >>>         {
#         >>>             'bucket'          : 'flukebook-prod-asset-store',
#         >>>             'key'             : '3/a/3a76b0e8-1c64-403d-ace1-679cf2f081c0/f2.jpg',
#         >>>         },
#         >>>     ],
#         >>>     'image_uuid_list': [
#         >>>         uuid.uuid4(),
#         >>>         uuid.uuid4(),
#         >>>         uuid.uuid4(),
#         >>>     ],
#         >>>     'image_width_list': [
#         >>>         1992,
#         >>>         1194,
#         >>>         500,
#         >>>     ],
#         >>>     'image_height_list': [
#         >>>         1328,
#         >>>         401,
#         >>>         500,
#         >>>     ],
#         >>> }
#         >>> gid_list = wbia.web.apis_json.add_images_json(web_instance, **_payload)
#         >>> print(gid_list)
#         >>> print(web_instance.get_image_uuids(gid_list))
#         >>> print(web_instance.get_image_uris(gid_list))
#         >>> print(web_instance.get_image_paths(gid_list))
#         >>> print(web_instance.get_image_uris_original(gid_list))
#     """
#     def _rectify(list_, default, length, func=None):
#         if list_ is None:
#             list_ = [None] * length

#         ret_list = []
#         for item in list_:

#             if item is None:
#                 item = default

#             if None not in [func, item]:
#                 item = func(item)

#             ret_list.append(item)

#         return ret_list

#     def _rectify_uri(list_, default, length, func=None):
#         list_ = _rectify(list_, default, length, func=None)

#         ret_list = []
#         for item in list_:

#             if isinstance(item, dict):
#                 item = ut.s3_dict_encode_to_str(item)

#             if ibs.containerized and item is not None:
#                 item = item.replace('://localhost/', '://nginx:80/')

#             ret_list.append(item)

#         return ret_list

#     def _verify(list_, tag, length, allow_none=False):
#         length_ = len(list_)
#         if length_ != length:
#             message = 'The input list %s has the wrong length. Received: %d. Expected %d'
#             args = (tag, length_, length, )
#             raise ValueError(message % args)

#         error_list = []
#         for value in enumerate(list_):
#             index, item = value
#             if item is None:
#                 error_list.append(value)

#         if len(error_list) > 0:
#             message = 'The input list %s has invalid values (index, value): %r'
#             args = (tag, error_list, )
#             raise ValueError(message % args)

#     def _uuid(value):
#         import uuid
#         import six

#         if value is None:
#             return None

#         if isinstance(value, six.string_types):
#             value = uuid.UUID(value)

#         return value

#     def _base(value):
#         if value is None:
#             return None

#         return basename(value)

#     def _ext(value):
#         if value is None:
#             return None

#         value = splitext(value)[1].lower()
#         value = '.jpg' if value == '.jpeg' else value
#         return value

#     # TODO: FIX ME SO THAT WE DON'T HAVE TO LOCALIZE EVERYTHING
#     kwargs['auto_localize'] = kwargs.get('auto_localize', True)
#     kwargs['sanitize'] = kwargs.get('sanitize', False)

#     expected_length = len(image_uri_list)

#     # Rectify values
#     image_uri_list         = _rectify_uri(image_uri_list    , None, expected_length,   str)
#     image_uuid_list        = _rectify(image_uuid_list       , None, expected_length, _uuid)
#     image_width_list       = _rectify(image_width_list      , None, expected_length,   int)
#     image_height_list      = _rectify(image_height_list     , None, expected_length,   int)
#     image_orig_name_list   = _rectify(image_uri_list        , None, expected_length, _base)
#     image_ext_list         = _rectify(image_uri_list        , None, expected_length,  _ext)
#     image_time_posix_list  = _rectify(image_time_posix_list ,   -1, expected_length, float)
#     image_gps_lat_list     = _rectify(image_gps_lat_list    , -1.0, expected_length, float)
#     image_gps_lon_list     = _rectify(image_gps_lon_list    , -1.0, expected_length, float)
#     image_orientation_list = _rectify(image_orientation_list,  0.0, expected_length,   int)
#     image_notes_list       = _rectify(image_notes_list      ,   '', expected_length,   str)

#     # Verify values
#     image_uri_list         = _verify(image_uri_list        , 'image_uri_list'        , expected_length)
#     image_uuid_list        = _verify(image_uuid_list       , 'image_uuid_list'       , expected_length)
#     image_width_list       = _verify(image_width_list      , 'image_width_list'      , expected_length)
#     image_height_list      = _verify(image_height_list     , 'image_height_list'     , expected_length)
#     image_orig_name_list   = _verify(image_orig_name_list  , 'image_orig_name_list'  , expected_length)
#     image_ext_list         = _verify(image_ext_list        , 'image_ext_list'        , expected_length)
#     image_time_posix_list  = _verify(image_time_posix_list , 'image_time_posix_list' , expected_length)
#     image_gps_lat_list     = _verify(image_gps_lat_list    , 'image_gps_lat_list'    , expected_length)
#     image_gps_lon_list     = _verify(image_gps_lon_list    , 'image_gps_lon_list'    , expected_length)
#     image_orientation_list = _verify(image_orientation_list, 'image_orientation_list', expected_length)
#     image_notes_list       = _verify(image_notes_list      , 'image_notes_list'      , expected_length)

#     params_gen = zip(
#         image_uuid_list,
#         image_uri_list,
#         image_uri_list,
#         image_orig_name_list,
#         image_ext_list,
#         image_width_list,
#         image_height_list,
#         image_time_posix_list,
#         image_gps_lat_list,
#         image_gps_lon_list,
#         image_orientation_list,
#         image_notes_list
#     )

#     gid_list = ibs.add_images(image_uri_list, params_list=params_gen, **kwargs)  # NOQA
#     image_uuid_list = ibs.get_image_uuids(gid_list)
#     return image_uuid_list


@register_api('/api/image/json/', methods=['POST'])
def add_images_json(
    ibs,
    image_uri_list,
    image_unixtime_list=None,
    image_gps_lat_list=None,
    image_gps_lon_list=None,
    **kwargs,
):
    """
    REST:
        Method: POST
        URL: /api/image/json/

    Ignore:
        sudo pip install boto

    Args:
        image_uri_list (list) : list of string image uris, most likely HTTP(S) or S3
            encoded URLs.  Alternatively, this can be a list of dictionaries (JSON
            objects) that specify AWS S3 stored assets.  An example below:

                image_uri_list = [
                    'http://domain.com/example/asset1.png',
                    '/home/example/Desktop/example/asset2.jpg',
                    's3://s3.amazon.com/example-bucket-2/asset1-in-bucket-2.tif',
                    {
                        'bucket'          : 'example-bucket-1',
                        'key'             : 'example/asset1.png',
                        'auth_domain'     : None,  # Uses 127.0.0.1
                        'auth_access_id'  : None,  # Uses system default
                        'auth_secret_key' : None,  # Uses system default
                    },
                    {
                        'bucket' : 'example-bucket-1',
                        'key'    : 'example/asset2.jpg',
                        # if unspecified, auth uses 127.0.0.1 and system defaults
                    },
                    {
                        'bucket'          : 'example-bucket-2',
                        'key'             : 'example/asset1-in-bucket-2.tif',
                        'auth_domain'     : 's3.amazon.com',
                        'auth_access_id'  : '____________________',
                        'auth_secret_key' : '________________________________________',
                    },
                ]

            Note that you cannot specify AWS authentication access ids or secret keys
            using string uri's.  For specific authentication methods, please use the
            latter list of dictionaries.

        image_time_posix_list (list of int): list of image's POSIX timestamps
        image_gps_lat_list (list of float): list of image's GPS latitude values
        image_gps_lon_list (list of float): list of image's GPS longitude values
        **kwargs : key-value pairs passed to the ibs.add_images() function.

    CommandLine:
        python -m wbia.web.apis_json --test-add_images_json


        ,"bucket":"flukebook-prod-asset-store","key":""

    Example:
        >>> # FIXME failing-test (03-Aug-2020) boto.exception.NoAuthHandlerFound: No handler was ready to authenticate
        >>> # xdoctest: +SKIP
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> import wbia
        >>> import uuid
        >>> web_instance = wbia.opendb(db='testdb1')
        >>> _payload = {
        >>>     'image_uri_list': [
        >>>         'https://upload.wikimedia.org/wikipedia/commons/4/49/Zebra_running_Ngorongoro.jpg',
        >>>         {
        >>>             'bucket'          : 'test-asset-store',
        >>>             'key'             : 'caribwhale/20130903-JAC-0002.JPG',
        >>>         },
        >>>         {
        >>>             'bucket'          : 'flukebook-prod-asset-store',
        >>>             'key'             : '3/a/3a76b0e8-1c64-403d-ace1-679cf2f081c0/f2.jpg',
        >>>         },
        >>>     ],
        >>> }
        >>> gid_list = wbia.web.apis_json.add_images_json(web_instance, **_payload)
        >>> print(gid_list)
        >>> print(web_instance.get_image_uuids(gid_list))
        >>> print(web_instance.get_image_uris(gid_list))
        >>> print(web_instance.get_image_paths(gid_list))
        >>> print(web_instance.get_image_uris_original(gid_list))
    """

    def _rectify(list_, default, length, func=None):
        if list_ is None:
            list_ = [None] * length

        ret_list = []
        for item in list_:

            if item is None:
                item = default

            if None not in [func, item]:
                item = func(item)

            ret_list.append(item)

        return ret_list

    def _rectify_uri(list_, default, length, func=str):
        list_ = _rectify(list_, default, length, func=None)

        ret_list = []
        for item in list_:

            if isinstance(item, dict):
                item = ut.s3_dict_encode_to_str(item)

            if ibs.containerized and item is not None:
                item = item.replace('://localhost/', '://nginx:80/')

            ret_list.append(item)

        return ret_list

    def _verify(list_, tag, length, allow_none=False):
        length_ = len(list_)
        if length_ != length:
            message = 'The input list %s has the wrong length. Received: %d. Expected %d'
            args = (
                tag,
                length_,
                length,
            )
            raise ValueError(message % args)

        error_list = []
        for value in enumerate(list_):
            index, item = value
            if item is None and not allow_none:
                error_list.append(value)

        if len(error_list) > 0:
            message = 'The input list %s has invalid values (index, value): %r'
            args = (
                tag,
                error_list,
            )
            raise ValueError(message % args)

        return list_

    kwargs['auto_localize'] = kwargs.get('auto_localize', True)
    kwargs['sanitize'] = kwargs.get('sanitize', False)

    depricated_list = [
        'image_uuid_list',
        'image_width_list',
        'image_height_list',
        'image_orig_name_list',
        'image_ext_list',
        'image_time_posix_list',
        'image_orientation_list',
        'image_notes_list',
    ]

    bad_list = []
    for depricated_value in depricated_list:
        if depricated_value in kwargs:
            bad_list.append(depricated_value)

    if len(bad_list) > 0:
        raise ValueError(
            'This API signature has changed, the following parameters have been deprecated: %r.  Please remove them and try again.'
            % (bad_list,)
        )

    expected_length = len(image_uri_list)

    # Rectify values
    image_uri_list = _rectify_uri(image_uri_list, None, expected_length, str)
    image_uri_list = _verify(image_uri_list, 'image_uri_list', expected_length)
    gid_list = ibs.add_images(image_uri_list, **kwargs)  # NOQA

    if image_unixtime_list is not None:
        image_unixtime_list = _rectify(image_unixtime_list, -1, expected_length, float)
        image_unixtime_list = _verify(
            image_unixtime_list, 'image_unixtime_list', expected_length, allow_none=True
        )

        flag_list = [
            None not in [gid, image_unixtime]
            for gid, image_unixtime in zip(gid_list, image_unixtime_list)
        ]
        gid_list_ = ut.filter_items(gid_list, flag_list)
        image_unixtime_list_ = ut.filter_items(image_unixtime_list, flag_list)

        logger.info('Setting times: %r -> %r' % (gid_list_, image_unixtime_list_))
        ibs.set_image_unixtime(gid_list_, image_unixtime_list_)

    if image_gps_lat_list is not None and image_gps_lon_list is not None:
        image_gps_lat_list = _rectify(image_gps_lat_list, -1.0, expected_length, float)
        image_gps_lon_list = _rectify(image_gps_lon_list, -1.0, expected_length, float)
        image_gps_lat_list = _verify(
            image_gps_lat_list, 'image_gps_lat_list', expected_length, allow_none=True
        )
        image_gps_lon_list = _verify(
            image_gps_lon_list, 'image_gps_lon_list', expected_length, allow_none=True
        )

        for index, value in enumerate(zip(image_gps_lat_list, image_gps_lon_list)):
            image_gps_lat, image_gps_lon = value
            if image_gps_lat is not None:
                assert (
                    image_gps_lon is not None
                ), 'Cannot specify a longitude without a latitude, index %d' % (index,)
            if image_gps_lon is not None:
                assert (
                    image_gps_lat is not None
                ), 'Cannot specify a longitude without a latitude, index %d' % (index,)

        flag_list = [
            None not in [gid, image_gps_lat_, image_gps_lon_]
            for gid, image_gps_lat_, image_gps_lon_ in zip(
                gid_list, image_gps_lat_list, image_gps_lon_list
            )
        ]
        gid_list_ = ut.filter_items(gid_list, flag_list)
        image_gps_lat_list_ = ut.filter_items(image_gps_lat_list, flag_list)
        image_gps_lon_list_ = ut.filter_items(image_gps_lon_list, flag_list)

        logger.info(
            'Setting gps: %r -> %r, %r'
            % (gid_list_, image_gps_lat_list_, image_gps_lon_list_)
        )
        ibs.set_image_gps(
            gid_list_, lat_list=image_gps_lat_list_, lon_list=image_gps_lon_list_
        )

    image_uuid_list = ibs.get_image_uuids(gid_list)
    return image_uuid_list


class ParseError(object):
    def __init__(self, value):
        self.value = value


@register_api('/api/annot/json/', methods=['POST'])
def add_annots_json(
    ibs,
    image_uuid_list,
    annot_bbox_list,
    annot_theta_list,
    annot_viewpoint_list=None,
    annot_quality_list=None,
    annot_species_list=None,
    annot_multiple_list=None,
    annot_interest_list=None,
    annot_name_list=None,
    **kwargs,
):
    """
    REST:
        Method: POST
        URL: /api/annot/json/

    Ignore:
        sudo pip install boto

    Args:
        image_uuid_list (list of str) : list of image UUIDs to be used in IBEIS IA
        annot_bbox_list (list of 4-tuple) : list of bounding box coordinates encoded as
            a 4-tuple of the values (xtl, ytl, width, height) where xtl is the
            'top left corner, x value' and ytl is the 'top left corner, y value'.
        annot_theta_list (list of float) : list of radian rotation around center.
            Defaults to 0.0 (no rotation).
        annot_species_list (list of str) : list of species for the annotation, if known.
            If the list is partially known, use None (null in JSON) for unknown entries.
        annot_name_list (list of str) : list of names for the annotation, if known.
            If the list is partially known, use None (null in JSON) for unknown entries.
        **kwargs : key-value pairs passed to the ibs.add_annots() function.

    CommandLine:
        python -m wbia.web.app --test-add_annots_json

    Example:
        >>> # DISABLE_DOCTEST
        >>> import wbia
        >>> from wbia.control.IBEISControl import *  # NOQA
        >>> web_instance = wbia.opendb(db='testdb1')
        >>> _payload = {
        >>>     'image_uuid_list': [
        >>>         uuid.UUID('7fea8101-7dec-44e3-bf5d-b8287fd231e2'),
        >>>         uuid.UUID('c081119a-e08e-4863-a710-3210171d27d6'),
        >>>     ],
        >>>     'annot_uuid_list': [
        >>>         uuid.UUID('fe1547c5-1425-4757-9b8f-b2b4a47f552d'),
        >>>         uuid.UUID('86d3959f-7167-4822-b99f-42d453a50745'),
        >>>     ],
        >>>     'annot_bbox_list': [
        >>>         [0, 0, 1992, 1328],
        >>>         [0, 0, 1194, 401],
        >>>     ],
        >>> }
        >>> aid_list = wbia.web.app.add_annots_json(web_instance, **_payload)
        >>> print(aid_list)
        >>> print(web_instance.get_annot_image_uuids(aid_list))
        >>> print(web_instance.get_annot_uuids(aid_list))
        >>> print(web_instance.get_annot_bboxes(aid_list))
    """

    def _rectify(list_, default, length, func=None):
        if list_ is None:
            list_ = [None] * length

        ret_list = []
        for item in list_:

            if item is None:
                item = default

            if None not in [func, item]:
                item = func(item)

            ret_list.append(item)

        return ret_list

    def _verify(list_, tag, length, allow_none=False):
        length_ = len(list_)
        if length_ != length:
            message = 'The input list %s has the wrong length. Received: %d. Expected %d'
            args = (
                tag,
                length_,
                length,
            )
            raise ValueError(message % args)

        error_list = []
        for value in enumerate(list_):
            index, item = value
            if item is None and not allow_none:
                error_list.append(value)
            if isinstance(item, ParseError):
                value = (index, item.value)
                error_list.append(value)

        if len(error_list) > 0:
            message = 'The input list %s has invalid values (index, value): %r'
            args = (
                tag,
                error_list,
            )
            raise ValueError(message % args)

        return list_

    def _uuid(value):
        if value is None:
            return ParseError(value)

        if isinstance(value, six.string_types):
            value = uuid.UUID(value)

        return value

    def _bbox(value):
        if len(value) != 4:
            return ParseError(value)

        value = tuple(map(float, value))

        return value

    depricated_list = [
        'annot_uuid_list',
        'annot_notes_list',
    ]

    bad_list = []
    for depricated_value in depricated_list:
        if depricated_value in kwargs:
            bad_list.append(depricated_value)

    if len(bad_list) > 0:
        raise ValueError(
            'This API signature has changed, the following parameters have been deprecated: %r.  Please remove them and try again.'
            % (bad_list,)
        )

    expected_length = len(image_uuid_list)

    image_uuid_list = _rectify(image_uuid_list, None, expected_length, _uuid)
    annot_bbox_list = _rectify(annot_bbox_list, None, expected_length, _bbox)
    annot_theta_list = _rectify(annot_theta_list, None, expected_length, float)

    image_uuid_list = _verify(image_uuid_list, 'image_uuid_list', expected_length)
    annot_bbox_list = _verify(annot_bbox_list, 'annot_bbox_list', expected_length)
    annot_theta_list = _verify(annot_theta_list, 'annot_theta_list', expected_length)

    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    gid_list = _verify(gid_list, 'image_uuid_list', expected_length)

    aid_list = ibs.add_annots(
        gid_list, bbox_list=annot_bbox_list, theta_list=annot_theta_list
    )

    if annot_viewpoint_list is not None:
        annot_viewpoint_list = _rectify(
            annot_viewpoint_list, const.VIEW.UNKNOWN, expected_length, str
        )
        annot_viewpoint_list = _verify(
            annot_viewpoint_list,
            'annot_viewpoint_list',
            expected_length,
            allow_none=True,
        )
        flag_list = [
            annot_viewpoint is not None for annot_viewpoint in annot_viewpoint_list
        ]
        aid_list_ = ut.filter_items(aid_list, flag_list)
        annot_viewpoint_list_ = ut.filter_items(annot_viewpoint_list, flag_list)
        ibs.set_annot_viewpoints(aid_list_, annot_viewpoint_list_)

    if annot_quality_list is not None:
        annot_quality_list = _rectify(
            annot_quality_list, const.QUAL_UNKNOWN, expected_length, str
        )
        annot_quality_list = _verify(
            annot_quality_list, 'annot_quality_list', expected_length, allow_none=True
        )
        flag_list = [annot_quality is not None for annot_quality in annot_quality_list]
        aid_list_ = ut.filter_items(aid_list, flag_list)
        annot_quality_list_ = ut.filter_items(annot_quality_list, flag_list)
        ibs.set_annot_quality_texts(aid_list_, annot_quality_list_)

    if annot_species_list is not None:
        annot_species_list = _rectify(
            annot_species_list, const.UNKNOWN, expected_length, str
        )
        annot_species_list = _verify(
            annot_species_list, 'annot_species_list', expected_length, allow_none=True
        )
        flag_list = [annot_species is not None for annot_species in annot_species_list]
        aid_list_ = ut.filter_items(aid_list, flag_list)
        annot_species_list_ = ut.filter_items(annot_species_list, flag_list)
        ibs.set_annot_species(aid_list_, annot_species_list_)

    if annot_multiple_list is not None:
        annot_multiple_list = _rectify(annot_multiple_list, False, expected_length, bool)
        annot_multiple_list = _verify(
            annot_multiple_list, 'annot_multiple_list', expected_length, allow_none=True
        )
        flag_list = [annot_multiple is not None for annot_multiple in annot_multiple_list]
        aid_list_ = ut.filter_items(aid_list, flag_list)
        annot_multiple_list_ = ut.filter_items(annot_multiple_list, flag_list)
        ibs.set_annot_multiple(aid_list_, annot_multiple_list_)

    if annot_interest_list is not None:
        annot_interest_list = _rectify(annot_interest_list, False, expected_length, bool)
        annot_interest_list = _verify(
            annot_interest_list, 'annot_interest_list', expected_length, allow_none=True
        )
        flag_list = [annot_interest is not None for annot_interest in annot_interest_list]
        aid_list_ = ut.filter_items(aid_list, flag_list)
        annot_interest_list_ = ut.filter_items(annot_interest_list, flag_list)
        ibs.set_annot_interest(aid_list_, annot_interest_list_)

    if annot_name_list is not None:
        annot_name_list = _rectify(annot_name_list, const.UNKNOWN, expected_length, str)
        annot_name_list = _verify(
            annot_name_list, 'annot_name_list', expected_length, allow_none=True
        )
        flag_list = [annot_name is not None for annot_name in annot_name_list]
        aid_list_ = ut.filter_items(aid_list, flag_list)
        annot_name_list_ = ut.filter_items(annot_name_list, flag_list)
        ibs.set_annot_names(aid_list_, annot_name_list_)

    annot_uuid_list = ibs.get_annot_uuids(aid_list)
    return annot_uuid_list


@register_api('/api/part/json/', methods=['POST'])
def add_parts_json(ibs, annot_uuid_list, part_bbox_list, part_theta_list, **kwargs):
    """
    REST:
        Method: POST
        URL: /api/part/json/

    Ignore:
        sudo pip install boto

    Args:
        annot_uuid_list (list of str) : list of annot UUIDs to be used in IBEIS IA
        part_uuid_list (list of str) : list of annotations UUIDs to be used in IBEIS IA
        part_bbox_list (list of 4-tuple) : list of bounding box coordinates encoded as
            a 4-tuple of the values (xtl, ytl, width, height) where xtl is the
            'top left corner, x value' and ytl is the 'top left corner, y value'.
        part_theta_list (list of float) : list of radian rotation around center.
            Defaults to 0.0 (no rotation).
        **kwargs : key-value pairs passed to the ibs.add_annots() function.
    """

    def _rectify(list_, default, length, func=None):
        if list_ is None:
            list_ = [None] * length

        ret_list = []
        for item in list_:

            if item is None:
                item = default

            if None not in [func, item]:
                item = func(item)

            ret_list.append(item)

        return ret_list

    def _verify(list_, tag, length, allow_none=False):
        length_ = len(list_)
        if length_ != length:
            message = 'The input list %s has the wrong length. Received: %d. Expected %d'
            args = (
                tag,
                length_,
                length,
            )
            raise ValueError(message % args)

        error_list = []
        for value in enumerate(list_):
            index, item = value
            if item is None and not allow_none:
                error_list.append(value)
            if isinstance(item, ParseError):
                value = (index, item.value)
                error_list.append(value)

        if len(error_list) > 0:
            message = 'The input list %s has invalid values (index, value): %r'
            args = (
                tag,
                error_list,
            )
            raise ValueError(message % args)

        return list_

    def _uuid(value):
        if value is None:
            return ParseError(value)

        if isinstance(value, six.string_types):
            value = uuid.UUID(value)

        return value

    def _bbox(value):
        if len(value) != 4:
            return ParseError(value)

        value = tuple(map(float, value))

        return value

    expected_length = len(annot_uuid_list)

    annot_uuid_list = _rectify(annot_uuid_list, None, expected_length, _uuid)
    part_bbox_list = _rectify(part_bbox_list, None, expected_length, _bbox)
    part_theta_list = _rectify(part_theta_list, None, expected_length, float)

    annot_uuid_list = _verify(annot_uuid_list, 'image_uuid_list', expected_length)
    part_bbox_list = _verify(part_bbox_list, 'part_bbox_list', expected_length)
    part_theta_list = _verify(part_theta_list, 'part_theta_list', expected_length)

    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    aid_list = _verify(aid_list, 'annot_uuid_list', expected_length)

    part_rowid_list = ibs.add_parts(
        aid_list, bbox_list=part_bbox_list, theta_list=part_theta_list
    )
    part_uuid_list = ibs.get_part_uuids(part_rowid_list)
    return part_uuid_list


@register_api('/api/name/json/', methods=['POST'])
def add_names_json(ibs, name_text_list, name_uuid_list=None, name_note_list=None):
    nid_list = ibs.add_names_json(
        name_text_list, name_uuid_list=name_uuid_list, name_note_list=name_note_list
    )
    return ibs.get_name_uuids(nid_list)


@register_api('/api/species/json/', methods=['POST'], __api_plural_check__=False)
def add_species_json(
    ibs,
    species_nice_list,
    species_text_list=None,
    species_code_list=None,
    species_uuid_list=None,
    species_note_list=None,
    skip_cleaning=False,
):
    species_rowid_list = ibs.add_species(
        species_nice_list,
        species_text_list=species_text_list,
        species_code_list=species_code_list,
        species_uuid_list=species_uuid_list,
        species_note_list=species_note_list,
        skip_cleaning=skip_cleaning,
    )
    return ibs.get_species_uuids(species_rowid_list)


@register_api('/api/match/json/', methods=['POST'])
def add_annotmatch_json(
    ibs,
    match_annot_uuid1_list,
    match_annot_uuid2_list,
    match_evidence_decision_list=None,
    match_meta_decision_list=None,
    match_confidence_list=None,
    match_user_list=None,
    match_tag_list=None,
    match_modified_list=None,
    match_count_list=None,
):
    aids1 = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aids2 = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)

    am_rowids = ibs.add_annotmatch_undirected(aids1, aids2)
    if match_evidence_decision_list:
        ibs.set_annotmatch_evidence_decision(am_rowids, match_evidence_decision_list)
    if match_meta_decision_list:
        ibs.set_annotmatch_meta_decision(am_rowids, match_meta_decision_list)
    if match_confidence_list:
        ibs.set_annotmatch_confidence(am_rowids, match_confidence_list)
    if match_user_list:
        ibs.set_annotmatch_reviewer(am_rowids, match_user_list)
    if match_tag_list:
        ibs.set_annotmatch_tag_text(am_rowids, match_tag_list)
    if match_modified_list:
        ibs.set_annotmatch_posixtime_modified(am_rowids, match_modified_list)
    if match_count_list:
        ibs.set_annotmatch_count(am_rowids, match_count_list)
    return list(zip(match_annot_uuid1_list, match_annot_uuid2_list))


@register_api('/api/review/json/', methods=['POST'])
def add_review_json(
    ibs,
    review_annot_uuid1_list,
    review_annot_uuid2_list,
    review_evidence_decision_list,
    review_meta_decision_list=None,
    review_uuid_list=None,
    review_user_list=None,
    review_user_confidence_list=None,
    review_tags_list=None,
    review_client_start_time_posix=None,
    review_client_end_time_posix=None,
    review_server_start_time_posix=None,
    review_server_end_time_posix=None,
):

    aids1 = ibs.get_annot_aids_from_uuid(review_annot_uuid1_list)
    aids2 = ibs.get_annot_aids_from_uuid(review_annot_uuid2_list)

    ibs.add_review(
        aids1,
        aids2,
        evidence_decision_list=review_evidence_decision_list,
        meta_decision_list=review_meta_decision_list,
        review_uuid_list=review_uuid_list,
        identity_list=review_user_list,
        user_confidence_list=review_user_confidence_list,
        tags_list=review_tags_list,
        review_client_start_time_posix=review_client_start_time_posix,
        review_client_end_time_posix=review_client_end_time_posix,
        review_server_start_time_posix=review_server_start_time_posix,
        review_server_end_time_posix=review_server_end_time_posix,
    )
    return list(zip(review_annot_uuid1_list, review_annot_uuid2_list))


@register_api('/api/imageset/json/', methods=['GET'])
def get_valid_imageset_uuids_json(ibs, **kwargs):
    imgsetid_list = ibs.get_valid_imgsetids(**kwargs)
    return ibs.get_imageset_uuid(imgsetid_list)


@register_api('/api/imageset/annot/uuid/json/', methods=['GET'])
def get_imageset_annot_uuids_json(ibs, imageset_uuid_list):
    imgsetid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    aids_list = ibs.get_imageset_aids(imgsetid_list)
    annot_uuids_list = [ibs.get_annot_uuids(aid_list) for aid_list in aids_list]
    return annot_uuids_list


@register_api('/api/imageset/num/annot/reviewed/json/', methods=['GET'])
def get_imageset_num_annots_reviewed_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_num_annots_reviewed(imageset_rowid_list)


@register_api('/api/imageset/num/image/reviewed/json/', methods=['GET'])
def get_imageset_num_imgs_reviewed_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_num_imgs_reviewed(imageset_rowid_list)


@register_api('/api/imageset/num/name/exemplar/json/', methods=['GET'])
def get_imageset_num_names_with_exemplar_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_num_names_with_exemplar(imageset_rowid_list)


@register_api('/api/imageset/num/image/json/', methods=['GET'])
def get_imageset_num_gids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_num_gids(imageset_rowid_list)


@register_api('/api/imageset/num/annot/json/', methods=['GET'])
def get_imageset_num_aids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_num_aids(imageset_rowid_list)


@register_api('/api/imageset/annot/rowid/json/', methods=['GET'])
def get_imageset_aids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_aids(imageset_rowid_list)


@register_api('/api/imageset/image/rowid/json/', methods=['GET'])
def get_imageset_gids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_gids(imageset_rowid_list)


@register_api('/api/imageset/image/uuid/json/', methods=['GET'])
def get_imageset_image_uuids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_image_uuids(imageset_rowid_list)


@register_api('/api/imageset/name/rowid/json/', methods=['GET'])
def get_imageset_nids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_nids(imageset_rowid_list)


@register_api('/api/imageset/name/uuid/json/', methods=['GET'])
def get_imageset_name_uuids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_name_uuids(imageset_rowid_list)


@register_api('/api/imageset/text/json/', methods=['GET'])
def get_imageset_text_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_text(imageset_rowid_list)


@register_api('/api/imageset/rowid/uuid/json/', methods=['GET'])
def get_imageset_imgsetids_from_uuid_json(ibs, imageset_uuid_list):
    return ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)


@register_api('/api/imageset/rowid/text/json/', methods=['GET'])
def get_imageset_imgsetids_from_text_json(ibs, imageset_text_list, **kwargs):
    return ibs.get_imageset_imgsetids_from_text(imageset_text_list, **kwargs)


@register_api('/api/imageset/note/json/', methods=['GET'])
def get_imageset_note_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_note(imageset_rowid_list)


@register_api('/api/imageset/time/posix/end/json/', methods=['GET'])
def get_imageset_end_time_posix_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_end_time_posix(imageset_rowid_list)


@register_api('/api/imageset/gps/lat/json/', methods=['GET'], __api_plural_check__=False)
def get_imageset_gps_lats_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_gps_lats(imageset_rowid_list)


@register_api('/api/imageset/gps/lon/json/', methods=['GET'], __api_plural_check__=False)
def get_imageset_gps_lons_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_gps_lons(imageset_rowid_list)


@register_api('/api/imageset/occurrence/json/', methods=['GET'])
def get_imageset_occurrence_flags_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_occurrence_flags(imageset_rowid_list)


@register_api('/api/imageset/processed/json/', methods=['GET'])
def get_imageset_processed_flags_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_processed_flags(imageset_rowid_list)


@register_api('/api/imageset/shipped/json/', methods=['GET'])
def get_imageset_shipped_flags_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_shipped_flags(imageset_rowid_list)


@register_api('/api/imageset/time/posix/start/json/', methods=['GET'])
def get_imageset_start_time_posix_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_start_time_posix(imageset_rowid_list)


@register_api('/api/imageset/duration/json/', methods=['GET'])
def get_imageset_duration_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_duration(imageset_rowid_list)


@register_api('/api/imageset/smart/waypoint/json/', methods=['GET'])
def get_imageset_smart_waypoint_ids_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_smart_waypoint_ids(imageset_rowid_list)


@register_api('/api/imageset/smart/xml/file/name/json/', methods=['GET'])
def get_imageset_smart_xml_fnames_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_smart_xml_fnames(imageset_rowid_list)


@register_api('/api/imageset/smart/xml/file/content/json/', methods=['GET'])
def get_imageset_smart_xml_contents_json(ibs, imageset_uuid_list):
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.get_imageset_smart_xml_contents(imageset_rowid_list)


@register_api('/api/image/json/<uuid>/', methods=['GET'])
def image_base64_api_json(ibs, uuid=None, thumbnail=False, fresh=False, **kwargs):
    rowid = ibs.get_image_rowid_from_uuid(uuid)
    return image_src(rowid, thumbnail=thumbnail, fresh=fresh, **kwargs)


@register_api('/api/image/json/', methods=['GET'])
def get_valid_image_uuids_json(ibs, **kwargs):
    gid_list = ibs.get_valid_gids(**kwargs)
    return ibs.get_image_uuids(gid_list)


@register_api('/api/image/dict/json/', methods=['GET'])
def get_image_uuids_with_annot_uuids(ibs, gid_list=None):
    if gid_list is None:
        gid_list = sorted(ibs.get_valid_gids())
    aids_list = ibs.get_image_aids(gid_list)
    zipped = list(zip(gid_list, aids_list))
    combined_dict = {
        str(ibs.get_image_uiids(gid)): ibs.get_annot_uuids(aid_list)
        for gid, aid_list in zipped
    }
    return combined_dict


@register_api('/api/image/rowid/uuid/json/', methods=['GET'])
def get_image_gids_from_uuid_json(ibs, image_uuid_list):
    return ibs.get_image_gids_from_uuid(image_uuid_list)


@register_api('/api/image/uri/json/', methods=['GET'])
def get_image_uris_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_uris(gid_list)


@register_api('/api/image/uri/original/json/', methods=['GET'])
def get_image_uris_original_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_uris_original(gid_list)


@register_api('/api/image/file/path/json/', methods=['GET'])
def get_image_paths_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_paths(gid_list)


@register_api('/api/image/file/hash/json/', methods=['GET'])
def get_image_hash_json(ibs, image_uuid_list, **kwargs):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_hash(gid_list, **kwargs)


@register_api('/api/image/file/name/json/', methods=['GET'])
def get_image_gnames_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_gnames(gid_list)


@register_api('/api/image/size/json/', methods=['GET'])
def get_image_sizes_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_sizes(gid_list)


@register_api('/api/image/width/json/', methods=['GET'])
def get_image_widths_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_widths(gid_list)


@register_api('/api/image/height/json/', methods=['GET'])
def get_image_heights_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_heights(gid_list)


@register_api('/api/image/gps/json/', methods=['GET'], __api_plural_check__=False)
def get_image_gps_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_gps(gid_list)


@register_api('/api/image/lat/json/', methods=['GET'])
def get_image_lat_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_lat(gid_list)


@register_api('/api/image/lon/json/', methods=['GET'])
def get_image_lon_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_lon(gid_list)


@register_api('/api/image/orientation/json/', methods=['GET'])
def get_image_orientation_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_orientation(gid_list)


@register_api('/api/image/orientation/str/json/', methods=['GET'])
def get_image_orientation_str_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_orientation_str(gid_list)


@register_api('/api/image/reviewed/json/', methods=['GET'])
def get_image_reviewed_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_reviewed(gid_list)


@register_api('/api/image/detect/confidence/json/', methods=['GET'])
def get_image_detect_confidence_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_detect_confidence(gid_list)


@register_api('/api/image/note/json/', methods=['GET'])
def get_image_notes_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_notes(gid_list)


@register_api('/api/image/name/rowid/json/', methods=['GET'])
def get_image_nids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_nids(gid_list)


@register_api('/api/image/name/uuid/json/', methods=['GET'])
def get_image_name_uuids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_name_uuids(gid_list)


@register_api(
    '/api/image/species/rowid/json/', methods=['GET'], __api_plural_check__=False
)
def get_image_species_rowids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_species_rowids(gid_list)


@register_api(
    '/api/image/species/uuid/json/', methods=['GET'], __api_plural_check__=False
)
def get_image_species_uuids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_species_uuids(gid_list)


@register_api('/api/image/imageset/rowid/json/', methods=['GET'])
def get_image_imgsetids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_imgsetids(gid_list)


@register_api('/api/image/imageset/uuid/json/', methods=['GET'])
def get_image_imgset_uuids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_imgset_uuids(gid_list)


@register_api('/api/image/imageset/text/json/', methods=['GET'])
def get_image_imagesettext_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_imagesettext(gid_list)


@register_api('/api/image/imageset/rowid/json/', methods=['PUT'])
def set_image_imgsetids_json(ibs, image_uuid_list, imageset_rowid_list):
    ibs.web_check_uuids(image_uuid_list, [], [])
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.set_image_imgsetids(gid_list, imageset_rowid_list)


@register_api('/api/image/imageset/uuid/json/', methods=['PUT'])
def set_image_imgset_uuids_json(ibs, image_uuid_list, imageset_uuid_list):
    ibs.web_check_uuids(image_uuid_list, [], [])
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    return ibs.set_image_imgsetids(gid_list, imageset_rowid_list)


@register_api('/api/image/imageset/text/json/', methods=['PUT'])
def set_image_imagesettext_json(ibs, image_uuid_list, imageset_text_list):
    ibs.web_check_uuids(image_uuid_list, [], [])
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    imageset_rowid_list = ibs.get_imageset_imgsetids_from_text(imageset_text_list)
    return ibs.set_image_imgsetids(gid_list, imageset_rowid_list)


@register_api('/api/image/annot/rowid/json/', methods=['GET'])
def get_image_aids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_aids(gid_list)


@register_api('/api/image/annot/uuid/json/', methods=['GET'])
def get_image_annot_uuids_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_annot_uuids(gid_list)


@register_api(
    '/api/image/annot/rowid/species/json/', methods=['GET'], __api_plural_check__=False
)
def get_image_aids_of_species_json(ibs, image_uuid_list, **kwargs):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_aids_of_species(gid_list, **kwargs)


@register_api(
    '/api/image/annot/uuid/species/json/', methods=['GET'], __api_plural_check__=False
)
def get_image_annot_uuids_of_species_json(ibs, image_uuid_list, **kwargs):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_annot_uuids_of_species(gid_list, **kwargs)


@register_api('/api/image/num/annot/json/', methods=['GET'])
def get_image_num_annotations_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_num_annotations(gid_list)


@register_api('/api/image/unixtime/json/', methods=['GET'])
def get_image_unixtimes_json(ibs, image_uuid_list):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_unixtime(gid_list)


@register_api('/api/image/timedelta/posix/json/', methods=['GET'])
def get_image_timedelta_posix_json(ibs, image_uuid_list, **kwargs):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_timedelta_posix(gid_list, **kwargs)


@register_api('/api/image/location/code/json/', methods=['GET'])
def get_image_location_codes_json(ibs, image_uuid_list, **kwargs):
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    return ibs.get_image_location_codes(gid_list, **kwargs)


@register_api('/api/annot/json/', methods=['GET'])
def get_valid_annot_uuids_json(ibs, **kwargs):
    aid_list = ibs.get_valid_aids(**kwargs)
    return ibs.get_annot_uuids(aid_list)


@register_api('/api/annot/json/<uuid>/', methods=['GET'])
def annotation_src_api_json(ibs, uuid=None):
    aid = ibs.get_annot_aids_from_uuid(uuid)
    return ibs.annotation_src_api(aid)


@register_api('/api/annot/rowid/uuid/json/', methods=['GET'])
def get_annot_aids_from_uuid_json(ibs, annot_uuid_list):
    return ibs.get_annot_aids_from_uuid(annot_uuid_list)


@register_api('/api/annot/image/rowid/json/', methods=['GET'])
def get_annot_gids_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_gids(aid_list)


@register_api('/api/annot/uuid/hashid/json/', methods=['GET'])
def get_annot_hashid_uuid_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_hashid_uuid(aid_list, **kwargs)


@register_api('/api/annot/exemplar/json/', methods=['POST'])
def set_exemplars_from_quality_and_viewpoint_json(
    ibs, annot_uuid_list, annot_name_list, **kwargs
):
    ibs.web_check_uuids([], annot_uuid_list, [])
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    if annot_name_list is not None:
        # Set names for query annotations
        nid_list = ibs.add_names(annot_name_list)
        ibs.set_annot_name_rowids(aid_list, nid_list)
    new_flag_list = ibs.set_exemplars_from_quality_and_viewpoint(aid_list, **kwargs)
    new_annot_uuid_list = ibs.get_annot_uuids(aid_list)
    return new_annot_uuid_list, new_flag_list


@register_api('/api/annot/bbox/json/', methods=['GET'])
def get_annot_bboxes_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_bboxes(aid_list)


@register_api('/api/annot/detect/confidence/json/', methods=['GET'])
def get_annot_detect_confidence_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_detect_confidence(aid_list)


@register_api('/api/annot/exemplar/json/', methods=['GET'])
def get_annot_exemplar_flags_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_exemplar_flags(aid_list)


@register_api('/api/annot/theta/json/', methods=['GET'])
def get_annot_thetas_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_thetas(aid_list)


@register_api('/api/annot/vert/json/', methods=['GET'])
def get_annot_verts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_verts(aid_list)


@register_api('/api/annot/vert/rotated/json/', methods=['GET'])
def get_annot_rotated_verts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_rotated_verts(aid_list)


@register_api('/api/annot/yaw/json/', methods=['GET'])
def get_annot_yaws_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_yaws(aid_list)


@register_api('/api/annot/viewpoint/json/', methods=['GET'])
def get_annot_viewpoints_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_viewpoints(aid_list)


@register_api('/api/annot/num/vert/json/', methods=['GET'])
def get_annot_num_verts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_num_verts(aid_list)


@register_api('/api/annot/name/rowid/json/', methods=['GET'])
def get_annot_nids_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_nids(aid_list, **kwargs)


@register_api('/api/annot/name/uuid/json/', methods=['GET'])
def get_annot_name_rowids_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    nid_list = ibs.get_annot_name_rowids(aid_list, **kwargs)
    return ibs.get_name_uuids(nid_list)


@register_api('/api/annot/name/text/json/', methods=['GET'])
def get_annot_name_texts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_name_texts(aid_list)


@register_api('/api/annot/note/json/', methods=['GET'])
def get_annot_notes_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_notes(aid_list)


@register_api('/api/annot/species/json/', methods=['GET'], __api_plural_check__=False)
def get_annot_species_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_species(aid_list)


@register_api(
    '/api/annot/species/rowid/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_species_rowids_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_species_rowids(aid_list)


@register_api(
    '/api/annot/species/uuid/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_species_uuids_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_species_uuids(aid_list)


@register_api(
    '/api/annot/species/text/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_species_texts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_species_texts(aid_list)


@register_api('/api/annot/imageset/rowid/json/', methods=['GET'])
def get_annot_imgsetids_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_imgsetids(aid_list)


@register_api('/api/annot/imageset/uuid/json/', methods=['GET'])
def get_annot_imgset_uuids_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_imgset_uuids(aid_list)


@register_api('/api/annot/imageset/text/json/', methods=['GET'])
def get_annot_image_set_texts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_image_set_texts(aid_list)


@register_api('/api/annot/image/name/json/', methods=['GET'])
def get_annot_image_names_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_image_names(aid_list)


@register_api('/api/annot/image/unixtime/json/', methods=['GET'])
def get_annot_image_unixtimes_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_image_unixtimes(aid_list)


@register_api('/api/annot/image/gps/json/', methods=['GET'], __api_plural_check__=False)
def get_annot_image_gps_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_image_gps(aid_list)


@register_api('/api/annot/image/file/path/json/', methods=['GET'])
def get_annot_image_paths_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_image_paths(aid_list)


@register_api('/api/annot/image/uuid/json/', methods=['GET'])
def get_annot_image_uuids_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    image_uuid_list = [
        None if aid is None else ibs.get_annot_image_uuids(aid) for aid in aid_list
    ]
    return image_uuid_list


@register_api('/api/annot/quality/json/', methods=['GET'])
def get_annot_qualities_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_qualities(aid_list, **kwargs)


@register_api('/api/annot/quality/text/json/', methods=['GET'])
def get_annot_quality_texts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_quality_texts(aid_list)


@register_api('/api/annot/yaw/text/json/', methods=['GET'])
def get_annot_yaw_texts_json(ibs, annot_uuid_list):
    # DEPRICATE
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_yaw_texts(aid_list)


@register_api('/api/annot/viewpoint/text/json/', methods=['GET'])
def get_annot_viewpoint_texts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_viewpoint_texts(aid_list)


@register_api('/api/annot/sex/json/', methods=['GET'])
def get_annot_sex_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_sex(aid_list, **kwargs)


@register_api('/api/annot/sex/text/json/', methods=['GET'])
def get_annot_sex_texts_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_sex_texts(aid_list)


@register_api('/api/annot/reviewed/json/', methods=['GET'])
def get_annot_reviewed_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_reviewed(aid_list)


@register_api('/api/annot/multiple/json/', methods=['GET'])
def get_annot_multiple_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_multiple(aid_list, **kwargs)


@register_api('/api/annot/interest/json/', methods=['GET'])
def get_annot_interest_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_interest(aid_list, **kwargs)


@register_api('/api/annot/image/contributor/tag/json/', methods=['GET'])
def get_annot_image_contributor_tag_json(ibs, annot_uuid_list):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_image_contributor_tag(aid_list)


@register_api('/api/annot/age/months/json/', methods=['GET'], __api_plural_check__=False)
def get_annot_age_months_est_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_age_months_est(aid_list, **kwargs)


@register_api(
    '/api/annot/age/months/text/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_age_months_est_texts_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_age_months_est_texts(aid_list, **kwargs)


@register_api(
    '/api/annot/age/months/min/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_age_months_est_min_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_age_months_est_min(aid_list, **kwargs)


@register_api(
    '/api/annot/age/months/max/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_age_months_est_max_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_age_months_est_max(aid_list, **kwargs)


@register_api(
    '/api/annot/age/months/min/text/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_age_months_est_min_texts_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_age_months_est_min_texts(aid_list, **kwargs)


@register_api(
    '/api/annot/age/months/max/text/json/', methods=['GET'], __api_plural_check__=False
)
def get_annot_age_months_est_max_texts_json(ibs, annot_uuid_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.get_annot_age_months_est_max_texts(aid_list, **kwargs)


@register_api('/api/annot/bbox/json/', methods=['PUT'])
def set_annot_bboxes_json(ibs, annot_uuid_list, bbox_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_bboxes(aid_list, bbox_list)


@register_api('/api/annot/theta/json/', methods=['PUT'])
def set_annot_thetas_json(ibs, annot_uuid_list, theta_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_thetas(aid_list, theta_list)


@register_api('/api/annot/viewpoint/json/', methods=['PUT'])
def set_annot_viewpoints_json(ibs, annot_uuid_list, viewpoint_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_viewpoints(aid_list, viewpoint_list)


@register_api('/api/annot/quality/text/json/', methods=['PUT'])
def set_annot_quality_texts_json(ibs, annot_uuid_list, quality_text_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_quality_texts(aid_list, quality_text_list)


@register_api('/api/annot/species/json/', methods=['PUT'], __api_plural_check__=False)
def set_annot_species_json(ibs, annot_uuid_list, species_text_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_species(aid_list, species_text_list)


@register_api('/api/annot/multiple/json/', methods=['PUT'])
def set_annot_multiple_json(ibs, annot_uuid_list, flag_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_multiple(aid_list, flag_list)


@register_api('/api/annot/interest/json/', methods=['PUT'])
def set_annot_interest_json(ibs, annot_uuid_list, flag_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_interest(aid_list, flag_list)


@register_api('/api/annot/name/text/json/', methods=['PUT'])
def set_annot_name_texts_json(ibs, annot_uuid_list, name_text_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    nid_list = ibs.get_name_rowids_from_text(name_text_list)
    return ibs.set_annot_name_rowids(aid_list, nid_list)


@register_api('/api/annot/note/json/', methods=['PUT'])
def set_annot_note_json(ibs, annot_uuid_list, annot_note_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_notes(aid_list, annot_note_list)


@register_api('/api/annot/tags/json/', methods=['PUT'], __api_plural_check__=False)
def set_annot_tag_text_json(ibs, annot_uuid_list, annot_tags_list, **kwargs):
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.set_annot_tag_text(aid_list, annot_tags_list)


@register_api('/api/name/json/', methods=['GET'])
def get_valid_name_uuids_json(ibs, **kwargs):
    nid_list = ibs.get_valid_nids(**kwargs)
    return ibs.get_name_uuids(nid_list)


@register_api('/api/name/dict/json/', methods=['GET'])
def get_name_nids_with_gids_json(ibs, nid_list=None):
    if nid_list is None:
        nid_list = sorted(ibs.get_valid_nids())
    name_list = ibs.get_name_texts(nid_list)
    gids_list = ibs.get_name_gids(nid_list)

    zipped = list(zip(nid_list, name_list, gids_list))
    combined_dict = {
        name: (ibs.get_name_uuids(nid), ibs.get_image_uuids(gid_list))
        for nid, name, gid_list in zipped
    }
    return combined_dict


@register_api('/api/name/annot/rowid/json/', methods=['GET'])
def get_name_aids_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_aids(nid_list, **kwargs)


@register_api('/api/name/annot/uuid/json/', methods=['GET'])
def get_name_annot_uuids_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_annot_uuids(nid_list, **kwargs)


@register_api('/api/name/annot/rowid/exemplar/json/', methods=['GET'])
def get_name_exemplar_aids_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_exemplar_aids(ibs, nid_list)


@register_api('/api/name/annot/uuid/exemplar/json/', methods=['GET'])
def get_name_exemplar_name_uuids_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_exemplar_name_uuids(nid_list, **kwargs)


@register_api('/api/name/image/rowid/json/', methods=['GET'])
def get_name_gids_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_gids(ibs, nid_list)


@register_api('/api/name/image/uuid/json/', methods=['GET'])
def get_name_image_uuids_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_image_uuids(ibs, nid_list)


@register_api('/api/name/note/json/', methods=['GET'])
def get_name_notes_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_notes(ibs, nid_list)


@register_api('/api/name/num/annot/json/', methods=['GET'])
def get_name_num_annotations_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_num_annotations(ibs, nid_list)


@register_api('/api/name/num/annot/exemplar/json/', methods=['GET'])
def get_name_num_exemplar_annotations_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_num_exemplar_annotations(ibs, nid_list)


@register_api('/api/name/temp/json/', methods=['GET'])
def get_name_temp_flag_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_temp_flag(nid_list, **kwargs)


@register_api('/api/name/alias/text/json/', methods=['GET'], __api_plural_check__=False)
def get_name_alias_texts_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_alias_texts(ibs, nid_list)


@register_api('/api/name/text/json/', methods=['GET'])
def get_name_texts_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_texts(nid_list, **kwargs)


@register_api('/api/name/uuid/text/json/', methods=['GET'])
def get_name_rowids_from_text_json(ibs, name_text_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_text(name_text_list, **kwargs)
    return ibs.get_name_uuids(nid_list)


@register_api('/api/name/rowid/uuid/json/', methods=['GET'])
def get_name_rowids_from_uuid_json(ibs, name_uuid_list, **kwargs):
    return ibs.get_name_rowids_from_uuid(name_uuid_list)


@register_api('/api/name/sex/json/', methods=['GET'])
def get_name_sex_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_sex(nid_list, **kwargs)


@register_api('/api/name/sex/text/json/', methods=['GET'])
def get_name_sex_text_json(ibs, name_uuid_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_sex_text(nid_list, **kwargs)


@register_api(
    '/api/name/age/months/min/json/', methods=['GET'], __api_plural_check__=False
)
def get_name_age_months_est_min_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_age_months_est_min(ibs, nid_list)


@register_api(
    '/api/name/age/months/max/json/', methods=['GET'], __api_plural_check__=False
)
def get_name_age_months_est_max_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_age_months_est_max(ibs, nid_list)


@register_api('/api/name/imageset/rowid/json/', methods=['GET'])
def get_name_imgsetids_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_imgsetids(ibs, nid_list)


@register_api('/api/name/imageset/uuid/json/', methods=['GET'])
def get_name_imgset_uuids_json(ibs, name_uuid_list):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.get_name_imgset_uuids(ibs, nid_list)


@register_api('/api/species/json/', methods=['GET'], __api_plural_check__=False)
def _get_all_species_rowids_json(ibs, **kwargs):
    species_rowid_list = ibs._get_all_species_rowids(**kwargs)
    return ibs.get_species_uuids(species_rowid_list, **kwargs)


@register_api(
    '/api/species/rowid/text/json/', methods=['GET'], __api_plural_check__=False
)
def get_species_rowids_from_text_json(ibs, species_text_list, **kwargs):
    return ibs.get_species_rowids_from_text(species_text_list, **kwargs)


@register_api(
    '/api/species/rowid/uuid/json/', methods=['GET'], __api_plural_check__=False
)
def get_species_rowids_from_uuids_json(ibs, species_uuid_list):
    return ibs.get_species_rowids_from_uuids(species_uuid_list)


@register_api('/api/species/text/json/', methods=['GET'], __api_plural_check__=False)
def get_species_texts_json(ibs, species_uuid_list):
    species_rowid_list = ibs.get_species_rowids_from_uuids(species_uuid_list)
    return ibs.get_species_texts(species_rowid_list)


@register_api('/api/species/nice/json/', methods=['GET'], __api_plural_check__=False)
def get_species_nice_json(ibs, species_uuid_list):
    species_rowid_list = ibs.get_species_rowids_from_uuids(species_uuid_list)
    return ibs.get_species_nice(species_rowid_list)


@register_api('/api/species/code/json/', methods=['GET'], __api_plural_check__=False)
def get_species_codes_json(ibs, species_uuid_list):
    species_rowid_list = ibs.get_species_rowids_from_uuids(species_uuid_list)
    return ibs.get_species_codes(species_rowid_list)


@register_api('/api/species/note/json/', methods=['GET'], __api_plural_check__=False)
def get_species_notes_json(ibs, species_uuid_list):
    species_rowid_list = ibs.get_species_rowids_from_uuids(species_uuid_list)
    return ibs.get_species_notes(species_rowid_list)


@register_api('/api/name/text/json/', methods=['PUT'])
def set_name_texts_json(ibs, name_uuid_list, name_text_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.set_name_texts(nid_list, name_text_list, **kwargs)


@register_api('/api/name/note/json/', methods=['PUT'])
def set_name_notes_json(ibs, name_uuid_list, name_note_list, **kwargs):
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    return ibs.set_name_notes(nid_list, name_note_list, **kwargs)


@register_api('/api/part/bbox/json/', methods=['PUT'])
def set_part_bboxes_json(ibs, part_uuid_list, bbox_list, **kwargs):
    aid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    return ibs.set_part_bboxes(aid_list, bbox_list)


@register_api('/api/part/theta/json/', methods=['PUT'])
def set_part_thetas_json(ibs, part_uuid_list, theta_list, **kwargs):
    aid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    return ibs.set_part_thetas(aid_list, theta_list)


@register_api('/api/part/viewpoint/json/', methods=['PUT'])
def set_part_viewpoints_json(ibs, part_uuid_list, viewpoint_list, **kwargs):
    aid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    return ibs.set_part_viewpoints(aid_list, viewpoint_list)


@register_api('/api/part/quality/text/json/', methods=['PUT'])
def set_part_quality_texts_json(ibs, part_uuid_list, quality_text_list, **kwargs):
    aid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    return ibs.set_part_quality_texts(aid_list, quality_text_list)


@register_api('/api/part/type/json/', methods=['PUT'], __api_plural_check__=False)
def set_part_types_json(ibs, part_uuid_list, type_text_list, **kwargs):
    aid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
    return ibs.set_part_types(aid_list, type_text_list)


# @register_api('/api/part/tags/json/', methods=['PUT'], __api_plural_check__=False)
# def set_part_tag_text_json(ibs, part_uuid_list, part_tags_list, **kwargs):
#     aid_list = ibs.get_part_rowids_from_uuid(part_uuid_list)
#     return ibs.set_part_tag_text(aid_list, part_tags_list)


@register_api('/api/match/decision/evidence/json/', methods=['PUT'])
def set_annotmatch_evidence_decision_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_decision_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_evidence_decision(
        annotmatch_rowid_list, match_decision_list
    )


@register_api('/api/match/decision/meta/json/', methods=['PUT'])
def set_annotmatch_meat_decision_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_decision_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_meta_decision(annotmatch_rowid_list, match_decision_list)


@register_api('/api/match/tags/json/', methods=['PUT'], __api_plural_check__=False)
def set_annotmatch_tag_text_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_tags_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_tag_text(annotmatch_rowid_list, match_tags_list)


@register_api('/api/match/confidence/json/', methods=['PUT'])
def set_annotmatch_confidence_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_confidence_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_confidence(annotmatch_rowid_list, match_confidence_list)


@register_api('/api/match/user/json/', methods=['PUT'])
def set_annotmatch_reviewer_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_user_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_reviewer(annotmatch_rowid_list, match_user_list)


@register_api('/api/match/count/json/', methods=['PUT'])
def set_annotmatch_count_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_count_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_count(annotmatch_rowid_list, match_count_list)


@register_api('/api/match/modified/json/', methods=['PUT'])
def set_annotmatch_posixtime_modified_json(
    ibs, match_annot_uuid1_list, match_annot_uuid2_list, match_modified_list, **kwargs
):
    aid1_list = ibs.get_annot_aids_from_uuid(match_annot_uuid1_list)
    aid2_list = ibs.get_annot_aids_from_uuid(match_annot_uuid2_list)
    annotmatch_rowid_list = ibs.get_annotmatch_rowid_from_superkey(aid1_list, aid2_list)
    return ibs.set_annotmatch_posixtime_modified(
        annotmatch_rowid_list, match_modified_list
    )


@register_api('/api/contributor/rowid/uuid/json/', methods=['GET'])
def get_contributor_rowids_from_uuid_json(ibs, contributor_uuid_list):
    return ibs.get_contributor_rowid_from_uuid(contributor_uuid_list)


@register_api('/api/part/json/', methods=['GET'])
def get_valid_part_uuids_json(ibs, **kwargs):
    part_rowid_list = ibs.get_valid_part_rowids(**kwargs)
    return ibs.get_part_uuids(part_rowid_list)


@register_api('/chaos/imageset/', methods=['GET', 'POST'], __api_plural_check__=False)
def chaos_imageset(ibs):
    """
    REST:
        Method: POST
        URL: /api/image/json/

    Args:
        image_uuid_list (list of str) : list of image UUIDs to be delete from IBEIS
    """
    from random import shuffle, randint

    gid_list = ibs.get_valid_gids()
    shuffle(gid_list)
    sample = min(len(gid_list) // 2, 50)
    assert sample > 0, 'Cannot create a chaos imageset using an empty database'
    gid_list_ = gid_list[:sample]
    imagetset_name = 'RANDOM_CHAOS_TEST_IMAGESET_%08d' % (randint(0, 99999999))
    imagetset_rowid = ibs.add_imagesets(imagetset_name)
    imagetset_uuid = ibs.get_imageset_uuid(imagetset_rowid)
    ibs.add_image_relationship(gid_list_, [imagetset_rowid] * len(gid_list_))
    return imagetset_name, imagetset_uuid


@register_api('/api/labeler/cnn/json/', methods=['POST'])
def labeler_cnn_json_wrapper(ibs, annot_uuid_list, **kwargs):
    ibs.web_check_uuids([], annot_uuid_list, [])
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    return ibs.labeler_cnn(aid_list, **kwargs)


@register_api('/api/imageset/json/', methods=['DELETE'])
def delete_imageset_json(ibs, imageset_uuid_list):
    """
    REST:
        Method: DELETE
        URL: /api/imageset/json/

    Args:
        imageset_uuid_list (list of str) : list of imageset UUIDs to be delete from IBEIS
    """
    imgsetid_list = ibs.get_imageset_imgsetids_from_uuid(imageset_uuid_list)
    ibs.delete_imagesets(imgsetid_list)
    return True


@register_api('/api/image/json/', methods=['DELETE'])
def delete_images_json(ibs, image_uuid_list):
    """
    REST:
        Method: DELETE
        URL: /api/image/json/

    Args:
        image_uuid_list (list of str) : list of image UUIDs to be delete from IBEIS
    """
    gid_list = ibs.get_image_gids_from_uuid(image_uuid_list)
    ibs.delete_images(gid_list)
    return True


@register_api('/api/annot/json/', methods=['DELETE'])
def delete_annots_json(ibs, annot_uuid_list):
    """
    REST:
        Method: DELETE
        URL: /api/annot/json/

    Args:
        annot_uuid_list (list of str) : list of annot UUIDs to be delete from IBEIS
    """
    aid_list = ibs.get_annot_aids_from_uuid(annot_uuid_list)
    ibs.delete_annots(aid_list)
    return True


@register_api('/api/species/json/', methods=['DELETE'], __api_plural_check__=False)
def delete_species_json(ibs, species_uuid_list):
    """
    REST:
        Method: DELETE
        URL: /api/species/json/

    Args:
        species_uuid_list (list of str) : list of species UUIDs to be delete from IBEIS
    """
    species_rowid_list = ibs.get_species_rowids_from_uuids(species_uuid_list)
    ibs.delete_species(species_rowid_list)
    return True


@register_api('/api/name/json/', methods=['DELETE'])
def delete_name_json(ibs, name_uuid_list):
    """
    REST:
        Method: DELETE
        URL: /api/name/json/

    Args:
        name_uuid_list (list of str) : list of name UUIDs to be delete from IBEIS
    """
    nid_list = ibs.get_name_rowids_from_uuid(name_uuid_list)
    ibs.delete_names(nid_list)
    return True



# -*- coding: utf-8 -*-
# flake8: noqa
import logging
import utool as ut
import numpy as np
import sklearn
import sklearn.datasets
import sklearn.svm
import sklearn.metrics
from sklearn import preprocessing

try:
    from wbia_cnn.models import abstract_models

    AbstractCategoricalModel = abstract_models.AbstractCategoricalModel
except ImportError:
    AbstractCategoricalModel = object
    print('no wbia_cnn')

from os.path import join

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def shark_net(dry=False):
    """
    CommandLine:
        python -m wbia.scripts.classify_shark shark_net
        python -m wbia.scripts.classify_shark shark_net --dry
        python -m wbia.scripts.classify_shark shark_net --vd --monitor

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.scripts.classify_shark import *  # NOQA
        >>> shark_net()
    """
    from wbia.scripts import classify_shark
    import wbia

    ibs = wbia.opendb('WS_ALL')  # NOQA
    config = {'dim_size': (224, 224), 'resize_dim': 'wh'}

    # ------------
    # Define dataset
    # ------------
    target_type = 'binary'
    # target_type = 'multiclass3'
    # ut.delete(ibs.get_neuralnet_dir())  # to reset
    dataset = classify_shark.get_shark_dataset(target_type, 'chip')

    # ------------
    # Define model
    # ------------
    if ut.get_computer_name() == 'Leviathan':
        batch_size = 128
        suffix = 'resnet'
        # suffix = 'lenet'
        # suffix = 'incep'
    else:
        suffix = 'lenet'
        batch_size = 64
        # suffix = 'resnet'
        # batch_size = 32

    model_name = 'injur-shark-' + suffix

    if False:
        model = classify_shark.WhaleSharkInjuryModel(
            name=model_name,
            output_dims=len(dataset.getprop('target_names')),
            data_shape=config['dim_size'] + (3,),
            batch_size=batch_size,
            arch_dpath='.',
        )
        model.init_arch()
        model.load_model_state()
    else:
        model = classify_shark.WhaleSharkInjuryModel(
            name=model_name,
            dataset_dpath=dataset.dataset_dpath,
            training_dpath=ibs.get_neuralnet_dir(),
            #
            output_dims=len(dataset.getprop('target_names')),
            data_shape=config['dim_size'] + (3,),
            batch_size=batch_size,
        )
        model.init_arch()
        model.print_layer_info()

    if False:
        model.arch_dpath = '/home/joncrall/Desktop/manually_saved/arch_injur-shark-resnet_o2_d27_c2942_jzuddodd/'

        state_fpath = model.get_model_state_fpath(dpath=model.trained_arch_dpath)
        state_fpath = model.get_model_state_fpath()
        model.load_model_state(fpath=state_fpath)

        # X_test, y_test = dataset.subset('test')
        # X_test, y_test = dataset.subset('valid')
        # X_test, y_test = dataset.subset('learn')
        X_test, y_test = dataset.subset('test')
        # y_pred = model.predict(X_test)
        test_outptuts = model._predict(X_test)
        y_pred = test_outptuts['predictions']
        logger.info(model.name)
        report = sklearn.metrics.classification_report(y_true=y_test, y_pred=y_pred)
        logger.info(report)

        state_fpath = '/home/joncrall/Desktop/manually_saved/arch_injur-shark-resnet_o2_d27_c2942_jzuddodd/model_state_arch_jzuddodd.pkl'
        dpath = '/home/joncrall/Desktop/manually_saved/arch_injur-shark-lenet_o2_d11_c688_acioqbst'
        model.dump_cases(X_test, y_test, 'test', dpath=dpath)

    hyperparams = dict(
        era_size=30,
        max_epochs=1000,
        rate_schedule=0.1,
        augment_on=True,
        class_weight='balanced',
        stopping_patience=200,
    )
    model.learn_state.weight_decay = 0.000002
    model.learn_state.learning_rate = 0.005
    ut.update_existing(model.hyperparams, hyperparams, assert_exists=True)
    model.monitor_config['monitor'] = True
    model.monitor_config['weight_dump_freq'] = 100
    model.monitor_config['case_dump_freq'] = 100

    # model.build_backprop_func()
    # model.build_forward_func()

    # ---------------
    # Setup and learn
    # ---------------

    X_learn, y_learn = dataset.subset('learn')
    X_valid, y_valid = dataset.subset('valid')
    X_test, y_test = dataset.subset('test')
    # model.ensure_data_params(X_learn, y_learn)
    # X_train = X_learn  # NOQA
    # y_train = y_learn  # NOQA
    valid_idx = None  # NOQA

    if dry or ut.get_argflag('--dry'):
        return model, dataset
    model.fit(X_learn, y_learn, X_valid=X_valid, y_valid=y_valid)


# @ut.reloadable_class
class WhaleSharkInjuryModel(AbstractCategoricalModel):
    """
    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.scripts.classify_shark import *  # NOQA
        >>> from wbia.scripts import classify_shark
        >>> ds = classify_shark.get_sharks_dataset('binary', 'chip')
        >>> problem = classify_shark.ClfProblem(ds)
        >>> problem.print_support_info()
        >>> ibs = ds.ibs
    """

    def def_lenet(model):
        import lasange
        from wbia_cnn import custom_layers

        logger.info('[model] init_arch')

        lrelu = lasange.nonlinearities.LeakyRectify(leakiness=(1.0 / 3.0))
        W = lasange.init.Orthogonal('relu')

        bundles = custom_layers.make_bundles(
            nonlinearity=lrelu,
            batch_norm=True,
            filter_size=(3, 3),
            stride=(1, 1),
            pool_size=(2, 2),
            pool_stride=(2, 2),
            W=W,
        )
        b = ut.DynStruct(copy_dict=bundles)

        network_layers_def = [
            b.InputBundle(shape=model.input_shape, noise=False),
            # Convolutional layers
            b.ConvBundle(num_filters=16, pool=True),
            b.ConvBundle(num_filters=16),
            b.ConvBundle(num_filters=16, pool=True),
            b.ConvBundle(num_filters=16),
            b.ConvBundle(num_filters=32, pool=True),
            b.ConvBundle(num_filters=32),
            b.ConvBundle(num_filters=32, pool=True),
            b.ConvBundle(num_filters=32),
            # Fully connected layers
            b.DenseBundle(num_units=64, dropout=0.5),
            b.DenseBundle(num_units=64, dropout=0.5),
            b.SoftmaxBundle(num_units=model.output_dims),
        ]
        return network_layers_def

    def def_resnet(model):
        import lasange
        from wbia_cnn import custom_layers

        logger.info('[model] init_arch')
        nonlinearity = lasange.nonlinearities.LeakyRectify(leakiness=(1.0 / 3.0))
        W = lasange.init.HeNormal(gain='relu')
        # W = lasange.init.GlorotUniform()

        bundles = custom_layers.make_bundles(
            nonlinearity=nonlinearity,
            filter_size=(3, 3),
            stride=(1, 1),
            W=W,
            pool_size=(2, 2),
            pool_stride=(2, 2),
        )
        b = ut.DynStruct(copy_dict=bundles)

        network_layers_def = [
            b.InputBundle(shape=model.input_shape, noise=False),
            # Convolutional layers
            b.ConvBundle(num_filters=16, pool=False),
            b.ResidualBundle(num_filters=16, stride=(2, 2), preactivate=False),
            b.ResidualBundle(num_filters=16),
            b.ResidualBundle(num_filters=16, stride=(2, 2)),
            b.ResidualBundle(num_filters=16),
            b.ResidualBundle(num_filters=16, stride=(2, 2)),
            b.ResidualBundle(num_filters=16),
            b.ResidualBundle(num_filters=16, stride=(2, 2)),
            b.ResidualBundle(num_filters=16, dropout=None),
            b.ResidualBundle(num_filters=16, stride=(2, 2), dropout=0.5),
            b.ResidualBundle(num_filters=16, postactivate=True, dropout=0.5),
            # Fully connected layers
            b.GlobalPool(),
            b.SoftmaxBundle(num_units=model.output_dims),
        ]
        return network_layers_def

    def def_inception(model):
        import lasange
        from wbia_cnn import custom_layers

        logger.info('[model] init_arch')

        N = 16

        # Define default incption branch types
        incep_branches = [
            dict(t='c', s=(1, 1), r=0, n=N),
            dict(t='c', s=(3, 3), r=N // 2, n=N // 2),
            dict(t='c', s=(3, 3), r=N // 4, n=N // 4, d=2),
            dict(t='p', s=(3, 3), n=N // 2),
        ]

        lrelu = lasange.nonlinearities.LeakyRectify(leakiness=(1.0 / 3.0))
        W = lasange.init.Orthogonal('relu')

        bundles = custom_layers.make_bundles(
            nonlinearity=lrelu,
            batch_norm=True,
            filter_size=(3, 3),
            stride=(1, 1),
            pool_size=(3, 3),
            pool_stride=(2, 2),
            branches=incep_branches,
            W=W,
        )
        b = ut.DynStruct(copy_dict=bundles)

        network_layers_def = [
            # Convolutional layers
            b.InputBundle(shape=model.input_shape, noise=False),
            b.ConvBundle(num_filters=16, filter_size=(3, 3), pool=False),
            b.ConvBundle(num_filters=12, filter_size=(3, 3), pool=True),
            b.InceptionBundle(dropout=0.3, pool=True),
            b.InceptionBundle(dropout=0.3, pool=True),
            b.InceptionBundle(dropout=0.4, pool=True),
            b.InceptionBundle(
                dropout=0.5,
                branches=[
                    dict(t='c', s=(1, 1), r=0, n=model.output_dims),
                    dict(t='c', s=(3, 3), r=N // 2, n=model.output_dims),
                    dict(t='c', s=(3, 3), r=N // 4, n=model.output_dims, d=2),
                    dict(t='p', s=(3, 3), n=model.output_dims),
                ],
            ),
            b.GlobalPool(),
            b.SoftmaxBundle(num_units=model.output_dims)
            # Fully connected layers
            # b.DenseBundle(num_units=64, dropout=.5),
            # b.DenseBundle(num_units=64, dropout=.5),
        ]
        return network_layers_def

    def init_arch(model, verbose=ut.VERBOSE, **kwargs):
        r"""

        CommandLine:
            python -m wbia.scripts.classify_shark WhaleSharkInjuryModel.init_arch
            python -m wbia.scripts.classify_shark WhaleSharkInjuryModel.init_arch --show

            python -m wbia.scripts.classify_shark shark_net --dry --show
            python -m wbia.scripts.classify_shark shark_net --vd

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.scripts.classify_shark import *  # NOQA
            >>> verbose = True
            >>> data_shape = tuple(ut.get_argval('--datashape', type_=list,
            >>>                                  default=(224, 224, 3)))
            >>> model = WhaleSharkInjuryModel(batch_size=64, output_dims=2,
            >>>                               data_shape=data_shape)
            >>> model.init_arch()
            >>> model.print_model_info_str()
            >>> ut.quit_if_noshow()
            >>> model.show_arch(fullinfo=False)
            >>> ut.show_if_requested()
        """
        from wbia_cnn import custom_layers

        # if ut.get_computer_name() == 'Leviathan':
        if model.name.endswith('incep'):
            network_layers_def = model.def_inception()
        elif model.name.endswith('lenet'):
            network_layers_def = model.def_lenet()
        elif model.name.endswith('resnet'):
            network_layers_def = model.def_resnet()
        network_layers = custom_layers.evaluate_layer_list(network_layers_def)
        # model.network_layers = network_layers
        output_layer = network_layers[-1]
        model.output_layer = output_layer
        return output_layer

    def special_output():
        pass

    # def special_loss_function(output_activations):
    #    output_injur1 = output_activations[:, 0]
    #    output_injur2 = output_activations[:, 1]
    #    output_healthy = (1 - ((1 - output_injur1) * (1 - output_injur2))
    #    import lasange
    #    lasange.objectives.binary_crossentropy(output_injur1)
    #    lasange.objectives.binary_crossentropy(output_injur2)

    def augment(self, Xb, yb=None):
        """
        X_valid, y_valid = dataset.subset('valid')
        num = 10
        Xb = X_valid[:num]
        Xb = Xb / 255.0 if ut.is_int(Xb) else Xb
        Xb = Xb.astype(np.float32, copy=True)
        yb = None if yb is None else yb.astype(np.int32, copy=True)
        # Rescale the batch data to the range 0 to 1
        Xb_, yb_ = model.augment(Xb)
        yb_ = None
        >>> ut.quit_if_noshow()
        >>> import wbia.plottool as pt
        >>> pt.qt4ensure()
        >>> from wbia_cnn import augment
        >>> augment.show_augmented_patches(Xb, Xb_, yb, yb_, data_per_label=1)
        >>> ut.show_if_requested()
        """
        from wbia_cnn import augment

        rng = np.random
        affperterb_ranges = dict(
            zoom_range=(1.3, 1.2),
            max_tx=2,
            max_ty=2,
            max_shear=ut.TAU / 32,
            max_theta=ut.TAU,
            enable_stretch=True,
            enable_flip=True,
        )
        Xb_, yb_ = augment.augment_affine(
            Xb,
            yb,
            rng=rng,
            inplace=True,
            data_per_label=1,
            affperterb_ranges=affperterb_ranges,
            aug_prop=0.5,
        )
        return Xb_, yb_

    # def fit_interactive(X_train, y_train, X_valid, y_valid):
    #    pass


def get_shark_dataset(target_type='binary', data_type='chip'):
    """
    >>> from wbia.scripts.classify_shark import *  # NOQA
    >>> target_type = 'binary'
    >>> data_type = 'hog'
    >>> dataset = get_shark_dataset(target_type)
    """
    from wbia_cnn.dataset import DataSet
    from wbia.scripts import classify_shark

    tup = classify_shark.get_shark_labels_and_metadata(target_type)
    ibs, annots, target, target_names, config, metadata, enc = tup
    data_shape = config['dim_size'] + (3,)
    length = len(annots)

    # Build dataset configuration string
    trail_cfgstr = ibs.depc_annot.get_config_trail_str('chips', config)
    trail_hashstr = ut.hashstr27(trail_cfgstr)
    visual_uuids = annots.visual_uuids
    metadata['visual_uuid'] = np.array(visual_uuids)
    # metadata['nids'] = np.array(annots.nids)
    chips_hashstr = ut.hashstr_arr27(annots.visual_uuids, 'chips')
    cfgstr = chips_hashstr + '_' + trail_hashstr
    name = 'injur-shark'

    if data_type == 'hog':
        cfgstr = 'hog_' + cfgstr
        name += '-hog'

    training_dpath = ibs.get_neuralnet_dir()
    dataset = DataSet(
        cfgstr,
        data_shape=data_shape,
        num_data=length,
        training_dpath=training_dpath,
        name=name,
    )

    logger.info(dataset.dataset_id)

    dataset.setprop('ibs', ibs)
    dataset.setprop('annots', annots)
    dataset.setprop('target_names', target_names)
    dataset.setprop('config', config)
    dataset.setprop('enc', enc)

    try:
        dataset.load()
    except IOError:
        import vtool as vt

        dataset.ensure_dirs()

        if data_type == 'hog':
            data = np.array([h.ravel() for h in annots.hog_hog])
            labels = target
            # Save data where dataset expects it to be
            dataset.save(data, labels, metadata, data_per_label=1)
        else:
            chip_gen = ibs.depc_annot.get(
                'chips', annots.aids, 'img', eager=False, config=config
            )
            iter_ = iter(ut.ProgIter(chip_gen, length=length, lbl='load chip'))
            shape = (length,) + data_shape
            data = vt.fromiter_nd(iter_, shape=shape, dtype=np.uint8)  # NOQA
            labels = target
            # Save data where dataset expects it to be
            dataset.save(data, labels, metadata, data_per_label=1)

    from wbia_cnn.dataset import stratified_label_shuffle_split

    if not dataset.has_split('learn'):
        nids = np.array(dataset.metadata['nids'])
        # Partition into a testing and training dataset
        y = dataset.labels
        train_idx, test_idx = stratified_label_shuffle_split(
            y, nids, [0.8, 0.2], rng=22019
        )
        nids_train = nids.take(train_idx, axis=0)
        y_train = y.take(train_idx, axis=0)
        # Partition training into learning and validation
        learn_idx, valid_idx = stratified_label_shuffle_split(
            y_train, nids_train, [0.8, 0.2], y_idx=train_idx, rng=90120
        )
        assert len(np.intersect1d(learn_idx, test_idx)) == 0
        assert len(np.intersect1d(valid_idx, test_idx)) == 0
        assert len(np.intersect1d(learn_idx, valid_idx)) == 0
        if data_type == 'hog':
            dataset.add_split('train', train_idx)
        dataset.add_split('test', test_idx)
        dataset.add_split('learn', learn_idx)
        dataset.add_split('valid', valid_idx)
        dataset.clear_cache('full')

    if data_type == 'hog':
        # hack
        y = dataset.labels
        nids = np.array(dataset.metadata['nids'])
        train_idx, test_idx = stratified_label_shuffle_split(
            y, nids, [0.8, 0.2], rng=22019
        )
        nids_train = nids.take(train_idx, axis=0)
        y_train = y.take(train_idx, axis=0)
        # Partition training into learning and validation
        learn_idx, valid_idx = stratified_label_shuffle_split(
            y_train, nids_train, [0.8, 0.2], y_idx=train_idx, rng=90120
        )
        dataset._split_idxs = {}
        dataset._split_idxs['learn'] = learn_idx
        dataset._split_idxs['valid'] = valid_idx
        dataset._split_idxs['train'] = train_idx
        dataset._split_idxs['test'] = test_idx

    dataset.ensure_symlinked()
    return dataset


def get_shark_labels_and_metadata(target_type=None, ibs=None, config=None):
    """
    >>> from wbia.scripts.classify_shark import *  # NOQA
    >>> target_type = 'multiclass3'
    >>> data_type = 'hog'
    """
    import wbia

    if ibs is None:
        ibs = wbia.opendb('WS_ALL')
    if config is None:
        config = {
            # 'dim_size': (256, 256),
            'dim_size': (224, 224),
            'resize_dim': 'wh',
        }
    all_annots = ibs.annots(config=config)

    isempty = ut.not_list(ut.lmap(len, ibs.images().aids))
    # if False:
    #    x = ibs.images().compress(isempty)
    num_empty_images = sum(isempty)
    logger.info('Images without annotations: %r' % (num_empty_images,))

    logger.info(
        'Building labels for %r annotations from %r images'
        % (len(all_annots), len(ut.unique(all_annots.gids)))
    )

    TARGET_TYPE = 'binary'
    # TARGET_TYPE = 'multiclass3'
    if target_type is None:
        target_type = TARGET_TYPE

    from wbia.scripts import getshark

    category_tags = getshark.get_injur_categories(all_annots)
    logger.info('Base Category Tags tags')
    logger.info(ut.repr3(ut.dict_hist(ut.flatten(category_tags))))

    logger.info('Base Co-Occurrence Freq')
    co_occur1 = ut.tag_coocurrence(category_tags)
    logger.info(ut.repr3(co_occur1))

    ntags_list = np.array(ut.lmap(len, category_tags))
    is_no_tag = ntags_list == 0
    is_single_tag = ntags_list == 1
    is_multi_tag = ntags_list > 1

    if target_type == 'binary':
        regex_map = [
            ('injur-.*', 'injured'),
            ('healthy', 'healthy'),
        ]
    elif target_type == 'multiclass3':
        regex_map = [
            ('injur-trunc', 'injur-trunc'),
            ('injur-nicks', 'injur-trunc'),
            ('injur-scar', 'injur-scar'),
            ('injur-bite', 'injur-scar'),
            ('injur-gill', 'injur-scar'),
            ('injur-other', None),
            ('injur-dead', None),
            ('healthy', 'healthy'),
        ]
    elif target_type == 'multiclassX':
        regex_map = [
            ('injur-trunc', 'injur-trunc'),
            ('healthy', 'healthy'),
            ('injur-.*', None),
        ]
    else:
        raise ValueError('Unknown target_type=%r' % (target_type,))

    tag_vocab = ut.flat_unique(*category_tags)
    alias_map = ut.build_alias_map(regex_map, tag_vocab)
    unmapped = list(set(tag_vocab) - set(alias_map.keys()))
    logger.info('unmapped = %r' % (unmapped,))
    category_tags2 = ut.alias_tags(category_tags, alias_map)

    ntags_list = np.array(ut.lmap(len, category_tags2))
    is_no_tag = ntags_list == 0
    is_single_tag = ntags_list == 1
    is_multi_tag = ntags_list > 1

    logger.info('Cleaned tags')
    hist = ut.tag_hist(category_tags2)
    logger.info(ut.repr3(hist))

    # Get tag co-occurrence
    logger.info('Co-Occurrence Freq')
    co_occur = ut.tag_coocurrence(category_tags2)
    logger.info(ut.repr3(co_occur))

    logger.info('Co-Occurrence Percent')
    co_occur_percent = ut.odict(
        [(keys, [100 * val / hist[k] for k in keys]) for keys, val in co_occur.items()]
    )
    logger.info(ut.repr3(co_occur_percent, precision=2, nl=1))

    multi_annots = all_annots.compress(is_multi_tag)  # NOQA
    # ibs.set_image_imagesettext(multi_annots.gids, ['MultiTaged'] * is_multi_tag.sum())

    logger.info("can't use %r annots due to no labels" % (is_no_tag.sum(),))
    logger.info("can't use %r annots due to inconsistent labels" % (is_multi_tag.sum(),))
    logger.info('will use %r annots with consistent labels' % (is_single_tag.sum(),))

    annot_tags = ut.compress(category_tags2, is_single_tag)
    annots = all_annots.compress(is_single_tag)
    annot_tag_hist = ut.dict_hist(ut.flatten(annot_tags))
    logger.info('Final Annot Tags')
    logger.info(ut.repr3(annot_tag_hist))

    # target_names = ['healthy', 'injured']
    enc = preprocessing.LabelEncoder()
    enc.fit(ut.unique(ut.flatten(annot_tags)))
    target = enc.transform(ut.flatten(annot_tags))
    target_names = enc.classes_

    metadata = {
        'aids': np.array(annots.aids),
        'nids': np.array(annots.nids),
    }
    tup = ibs, annots, target, target_names, config, metadata, enc
    return tup


# @ut.reloadable_class
class ClfProblem(object):
    """ Harness for researching a classification problem """

    def __init__(problem, ds):
        problem.ds = ds

    def print_support_info(problem):
        enc = problem.ds.enc
        target_labels = enc.inverse_transform(problem.ds.target)
        label_hist = ut.dict_hist(target_labels)
        logger.info('support hist' + ut.repr3(label_hist))

    def fit_new_classifier(problem, train_idx):
        """
        References:
            http://leon.bottou.org/research/stochastic
            http://blog.explainmydata.com/2012/06/ntrain-24853-ntest-25147-ncorrupt.html
            http://scikit-learn.org/stable/modules/svm.html#svm-classification
            http://scikit-learn.org/stable/modules/grid_search.html
        """
        logger.info('[problem] train classifier on %d data points' % (len(train_idx)))
        data = problem.ds.data
        target = problem.ds.target
        x_train = data.take(train_idx, axis=0)
        y_train = target.take(train_idx, axis=0)
        clf = sklearn.svm.SVC(
            kernel=str('linear'),
            C=0.17,
            class_weight='balanced',
            decision_function_shape='ovr',
        )

        # C, penalty, loss
        # param_grid = {'C': [1e3, 5e3, 1e4, 5e4, 1e5],
        #              'gamma': [0.0001, 0.0005, 0.001, 0.005, 0.01, 0.1], }
        # param_grid = {'C': [1e3, 5e3, 1e4, 5e4, 1e5],
        #              'gamma': [0.0001, 0.0005, 0.001, 0.005, 0.01, 0.1], }
        # clf = GridSearchCV(SVC(kernel='rbf', class_weight='balanced'), param_grid)
        # clf = clf.fit(X_train_pca, y_train)
        clf.fit(x_train, y_train)
        return clf

    def fit_new_linear_svm(problem, train_idx):
        logger.info('[problem] train classifier on %d data points' % (len(train_idx)))
        data = problem.ds.data
        target = problem.ds.target
        x_train = data.take(train_idx, axis=0)
        y_train = target.take(train_idx, axis=0)
        clf = sklearn.svm.SVC(
            kernel=str('linear'),
            C=0.17,
            class_weight='balanced',
            decision_function_shape='ovr',
        )
        clf.fit(x_train, y_train)

    def gridsearch_linear_svm_params(problem, train_idx):
        """
        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.scripts.classify_shark import *  # NOQA
            >>> from wbia.scripts import classify_shark
            >>> ds = classify_shark.get_sharks_dataset('binary')
            >>> problem = classify_shark.ClfProblem(ds)
            >>> problem.print_support_info()
        """
        try:
            import sklearn.model_selection
        except ImportError:
            pass
        import sklearn.grid_search

        with ut.Timer('cv'):
            data = problem.ds.data
            target = problem.ds.target

            def stratified_sample_idxs_unbalanced(target, size=1000):
                rng = np.random.RandomState(43)
                sample = []
                for label in np.unique(target):
                    target_idxs = np.where(target == label)[0]
                    subset_size = size
                    rand_idx = ut.random_indexes(len(target_idxs), subset_size, rng=rng)
                    sample_idx = ut.take(target_idxs, rand_idx)
                    sample.append(sample_idx)
                sample_idx = np.array(sorted(ut.flatten(sample)))
                return sample_idx

            train_idx = stratified_sample_idxs_unbalanced(target, 4000)

            x_train = data.take(train_idx, axis=0)
            y_train = target.take(train_idx, axis=0)
            param_grid = {
                # 'C': [1, .5, .1, 5, 10, 100],
                # 'C': [1, 1e-1, 1e-2, 1e-3]
                # 'C': [1, 1e-1, 1e-2, 1e-3]
                # 'C': np.linspace(1, 1e-5, 15)
                # 'C': np.linspace(.2, 1e-5, 15)
                # 'C': np.logspace(np.log10(1e-3), np.log10(.1), 30, base=10)
                # 'C': np.linspace(.1, .3, 20),
                # 'C': np.linspace(1.0, .22, 20),
                'C': np.linspace(0.25, 0.01, 40),
                # 'loss': ['l2', 'l1'],
                # 'penalty': ['l2', 'l1'],
            }
            _clf = sklearn.svm.SVC(
                kernel=str('linear'),
                C=0.17,
                class_weight='balanced',
                decision_function_shape='ovr',
            )
            clf = sklearn.grid_search.GridSearchCV(
                _clf, param_grid, n_jobs=6, iid=False, cv=5, verbose=3
            )
            clf.fit(x_train, y_train)
            # (NOTE grid.predict only uses the best estimator)
            logger.info('clf.best_params_ = %r' % (clf.best_params_,))
            logger.info('Best parameters set found on development set:')
            logger.info(clf.best_params_)
            logger.info('Grid scores on development set:')
            for params, mean_score, scores in clf.grid_scores_:
                logger.info(
                    '%0.3f (+/-%0.03f) for %r' % (mean_score, scores.std() * 2, params)
                )
            xdata = np.array([t[0]['C'] for t in clf.grid_scores_])
            ydata = np.array([t[1] for t in clf.grid_scores_])

            y_data_std = np.array([t[2].std() for t in clf.grid_scores_])
            ydata_mean = ydata
            y_data_max = ydata_mean + y_data_std
            y_data_min = ydata_mean - y_data_std

            # pt.plot(xdata, ydata, '-rx')
            import wbia.plottool as pt

            pt.figure(fnum=pt.ensure_fnum(None))
            ax = pt.gca()
            ax.fill_between(xdata, y_data_min, y_data_max, alpha=0.2, color=pt.LIGHT_BLUE)
            pt.draw_hist_subbin_maxima(ydata, xdata)

            # y_data_std = np.array([t[2].std() for t in grid.grid_scores_])
            # ydata_mean = c_ydata
            # y_data_max = ydata_mean + y_data_std
            # y_data_min = ydata_mean - y_data_std
            # # import wbia.plottool as pt
            # # pt.figure(fnum=pt.ensure_fnum(None))
            # ax = pt.gca()
            # ax.fill_between(c_xdata, c_ydata, y_data_max, alpha=.2, color=pt.LIGHT_BLUE)
            # ax.fill_between(c_xdata, c_ydata, y_data_min, alpha=.2, color=pt.LIGHT_BLUE)
            # # pt.figure(fnum=pt.ensure_fnum(None))
            # hist = c_ydata
            # centers = c_xdata
            # pt.draw_hist_subbin_maxima(c_ydata, c_xdata, maxima_thresh=None, remove_endpoints=False)

            # clf.best_params_ = {u'C': 0.07143785714285722}
            # Best parameters set found on development set:
            # {u'C': 0.07143785714285722}
            # Grid scores on development set:
            # 0.729 (+/-0.016) for {u'C': 1.0}
            # 0.729 (+/-0.019) for {u'C': 0.92857214285714285}
            # 0.733 (+/-0.017) for {u'C': 0.85714428571428569}
            # 0.734 (+/-0.015) for {u'C': 0.78571642857142865}
            # 0.736 (+/-0.016) for {u'C': 0.71428857142857138}
            # 0.739 (+/-0.020) for {u'C': 0.64286071428571434}
            # 0.742 (+/-0.020) for {u'C': 0.57143285714285719}
            # 0.743 (+/-0.021) for {u'C': 0.50000500000000003}
            # 0.746 (+/-0.023) for {u'C': 0.42857714285714288}
            # 0.749 (+/-0.023) for {u'C': 0.35714928571428572}
            # 0.755 (+/-0.025) for {u'C': 0.28572142857142857}
            # 0.760 (+/-0.027) for {u'C': 0.21429357142857142}
            # 0.762 (+/-0.025) for {u'C': 0.14286571428571437}
            # 0.770 (+/-0.036) for {u'C': 0.07143785714285722}
            # 0.664 (+/-0.031) for {u'C': 1.0000000000000001e-05}

            # 0.774 (+/-0.039) for {u'C': 0.017433288221999882}
            # 0.775 (+/-0.039) for {u'C': 0.020433597178569417}
            # 0.774 (+/-0.039) for {u'C': 0.023950266199874861}
            # 0.777 (+/-0.038) for {u'C': 0.02807216203941177}
            # 0.775 (+/-0.036) for {u'C': 0.032903445623126679}
            # 0.773 (+/-0.033) for {u'C': 0.038566204211634723}

            # 0.722 (+/-0.060) for {u'C': 0.001}
            # 0.770 (+/-0.047) for {u'C': 0.01}
            # 0.775 (+/-0.047) for {u'C': 0.1}
            # 0.774 (+/-0.047) for {u'C': 0.12}
            # 0.773 (+/-0.045) for {u'C': 0.15}
            # 0.773 (+/-0.046) for {u'C': 0.17}
            # 0.772 (+/-0.047) for {u'C': 0.2}
            # 0.760 (+/-0.043) for {u'C': 0.5}
            # 0.748 (+/-0.043) for {u'C': 1.0}
            # 0.707 (+/-0.043) for {u'C': 100}
            # 0.702 (+/-0.047) for {u'C': 1000}

    def classifier_test(problem, clf, test_idx):
        logger.info('[problem] test classifier on %d data points' % (len(test_idx),))
        data = problem.ds.data
        target = problem.ds.target
        X_test = data.take(test_idx, axis=0)
        y_true = target.take(test_idx, axis=0)

        y_conf = predict_svc_ovr(X_test)
        y_pred = y_conf.argmax(axis=1)

        result = ClfSingleResult(problem.ds, test_idx, y_true, y_pred, y_conf)
        return result

    def stratified_2sample_idxs(problem, frac=0.2, split_frac=0.75):
        target = problem.ds.target
        target_labels = problem.ds.target_labels

        rng = np.random.RandomState(43)
        train_sample = []
        test_sample = []
        for label in target_labels:
            target_idxs = np.where(target == label)[0]
            subset_size = int(len(target_idxs) * frac)
            rand_idx = ut.random_indexes(len(target_idxs), subset_size, rng=rng)
            sample_idx = ut.take(target_idxs, rand_idx)
            split = int(len(sample_idx) * split_frac)
            train_sample.append(sample_idx[split:])
            test_sample.append(sample_idx[:split])

        train_idx = np.array(sorted(ut.flatten(train_sample)))
        test_idx = np.array(sorted(ut.flatten(test_sample)))
        return train_idx, test_idx

    def gen_crossval_idxs(problem, n_folds=2):
        y = problem.ds.target
        rng = 43432
        if hasattr(problem.ds, 'nids'):
            # Ensure that an individual does not appear in both the train
            # and the test dataset
            from wbia_cnn.dataset import stratified_kfold_label_split

            labels = problem.ds.nids
            _iter = stratified_kfold_label_split(y, labels, n_folds=n_folds, rng=rng)
        else:
            xvalkw = dict(n_folds=n_folds, shuffle=True, random_state=rng)
            import sklearn.cross_validation

            skf = sklearn.cross_validation.StratifiedKFold(y, **xvalkw)
            _iter = skf
            # import sklearn.model_selection
            # skf = sklearn.model_selection.StratifiedKFold(**xvalkw)
            # _iter = skf.split(X=np.empty(len(y)), y=y)
        msg = 'cross-val test on %s' % (problem.ds.name)
        progiter = ut.ProgIter(_iter, length=n_folds, lbl=msg)
        for train_idx, test_idx in progiter:
            yield train_idx, test_idx


# @ut.reloadable_class
class ClfSingleResult(object):
    r"""
    Reports the results of a classification problem

    Example:
        >>> # DISABLE_DOCTEST
        >>> result = ClfSingleResult()
    """

    def __init__(result, ds=None, test_idx=None, y_true=None, y_pred=None, y_conf=None):
        result.ds = ds
        result.test_idx = test_idx
        result.y_true = y_true
        result.y_pred = y_pred
        result.y_conf = y_conf

    def compile_results(result):
        import pandas as pd

        y_true = result.y_true
        y_pred = result.y_pred
        y_conf = result.y_conf
        test_idx = result.test_idx

        index = pd.Series(test_idx, name='test_idx')
        if len(result.ds.target_names) == 1:
            y_conf
        decision = pd.DataFrame(y_conf, index=index, columns=result.ds.target_names)
        result.decision = decision / 3
        easiness = np.array(ut.ziptake(result.decision.values, y_true))
        columns = ['pred', 'easiness']
        column_data = [y_pred, easiness]
        data = dict(zip(columns, column_data))
        result.df = pd.DataFrame(data, index, columns)

    def print_report(result):
        report = sklearn.metrics.classification_report(
            result.y_true, result.y_pred, target_names=result.ds.target_names
        )
        logger.info(report)


def get_model_state(clf):
    model_attr_names = [a for a in dir(clf) if a.endswith('_') and not a.startswith('__')]
    model_state = {a: getattr(clf, a) for a in model_attr_names}
    return model_state


def set_model_state(clf, model_state):
    attr_names = sorted(model_state.keys())
    attr_names1 = [
        'dual_coef_',
    ]
    attr_names2 = [
        'coef_',
    ]
    attr_names3 = attr_names1 + attr_names2
    attr_namesA = ut.isect(attr_names1, attr_names)
    attr_namesB = ut.setdiff(attr_names, attr_names3)
    attr_namesC = ut.isect(attr_names2, attr_names)
    attr_names_ = attr_namesA + attr_namesB + attr_namesC
    for a in attr_names_:
        val = model_state[a]
        logger.info('a = %r' % (a,))
        try:
            setattr(clf, a, val)
        except AttributeError:
            val2 = getattr(clf, a)
            assert np.all(val == val2)


def predict_svc_ovr(clf, data):
    if len(clf.classes_) == 2:
        X = clf._validate_for_predict(data)
        X = clf._compute_kernel(X)
        _dec2 = clf._dense_decision_function(X)
        dec2 = -_dec2

        n_samples = dec2.shape[0]
        n_classes = len(clf.classes_)
        final = np.zeros((n_samples, n_classes))
        confidence_max = max(np.abs(dec2.max()), np.abs(dec2.min()))
        norm_conf = ((dec2.T[0] / confidence_max) + 1) / 2
        final.T[0] = 1 - norm_conf
        final.T[1] = norm_conf
        # output comparable to multiclass version
        y_conf = final
    else:
        # Get notion of confidence / probability of decision
        y_conf = clf.decision_function(data)
    return y_conf


def predict_ws_injury_interim_svm(ibs, aids, **kwargs):
    """
    Returns relative confidence
    """
    config = {
        # 'dim_size': (256, 256),
        'dim_size': (224, 224),
        'resize_dim': 'wh',
    }

    # Load the SVM
    model_fname = 'interim_svc_injur-shark-hog_12559_224x224x3_ldhhxnxo.cPkl'
    model_url = 'https://wildbookiarepository.azureedge.net/models/{}'.format(model_fname)
    model_fpath = ut.grab_file_url(model_url, check_hash=False)
    clf = ut.load_cPkl(model_fpath)

    annots = ibs.annots(aids, config=config)
    data = np.array([h.ravel() for h in annots.hog_hog])

    target_names = ['healthy', 'injured']
    # confidence = clf.decision_function(data)
    # y_conf = predict_svc_ovr(clf, data)
    scores = clf.decision_function(data)
    y_pred = scores > 0.0
    y_pred = y_pred.astype(np.int64)
    # y_pred = clf.predict(data)
    ut.embed()
    pred_nice = ut.take(target_names, y_pred)
    return list(zip(pred_nice, scores))


def shark_svm():
    r"""
    References:
        http://scikit-learn.org/stable/model_selection.html

    TODO:
        * Change unreviewed healthy tags to healthy-likely

    CommandLine:
        python -m wbia.scripts.classify_shark shark_svm --show
        python -m wbia.scripts.classify_shark shark_svm

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.scripts.classify_shark import *  # NOQA
        >>> shark_svm()
        >>> ut.show_if_requested()
    """
    from wbia.scripts import classify_shark

    target_type = 'binary'
    # target_type = 'multiclass3'
    # dataset = classify_shark.get_shark_dataset(target_type)

    ds = classify_shark.get_shark_dataset(target_type, 'hog')
    # Make resemble old dataset
    # FIXME; make wbia_cnn dataset work here too
    # annots = ds.getprop('annots')
    ds.enc = ds.getprop('enc')
    ds.aids = ds.getprop('annots').aids
    ds.nids = ds.getprop('annots').nids
    ds.target = ds.labels
    ds.target_names = ds.getprop('target_names')
    ds.target_labels = ds.enc.transform(ds.target_names)
    ds.ibs = ds.getprop('ibs')
    ds.config = ds.getprop('config')

    problem = classify_shark.ClfProblem(ds)
    problem.print_support_info()

    # BUILD_RELEASE_MODEL = False
    # if BUILD_RELEASE_MODEL:
    #     clf = sklearn.svm.SVC(kernel=str('linear'), C=.17,
    #                           class_weight='balanced',
    #                           decision_function_shape='ovr',
    #                           verbose=10)
    #     clf.fit(ds.data, ds.target)
    #     model_fname = 'interim_svc_{}.cPkl'.format(ds.dataset_id)
    #     model_dpath = ut.ensuredir((ds.dataset_dpath, 'svms'))
    #     model_fpath = join(model_dpath, model_fname)
    #     ut.save_cPkl(model_fpath, clf)
    #     """
    #     TO PUBLISH
    #     scp clf to lev:/media/hdd/PUBLIC/models
    #     run script lev:/media/hdd/PUBLIC/hash.py to refresh hashes
    #     """
    #     user = ut.get_user_name()
    #     host = 'cthulhu.dyn.wildme.io'
    #     remote_path = '/data/public/models/' + model_fname
    #     remote_uri = user + '@' + host + ':' + remote_path
    #     ut.rsync(model_fpath, remote_uri)

    #     command = 'python /media/hdd/PUBLIC/hash.py'
    #     ut.cmd('ssh {user}@{host} "{command}"'.format(user=user, host=host,
    #                                                   command=command))

    model_dpath = ut.ensuredir((ds.dataset_dpath, 'svms'))
    # n_folds = 10
    n_folds = 10
    # ensemble_dpath = ut.ensuredir((model_dpath, 'svms_%d_fold' % (n_folds,)))

    train_idx = ds._split_idxs['train']
    test_idx = ds._split_idxs['test']

    y_train = ds.target.take(train_idx)
    nids_train = ut.take(ds.nids, train_idx)

    # Ensure that an individual does not appear in both train and test
    # _iter = stratified_kfold_label_split(y_train, nids_train, y_idx=train_idx,
    #                                     n_folds=n_folds, rng=rng)

    class MyLabelCV(object):
        def __init__(self, y_train, nids_train, n_folds):
            self.nids_train = nids_train
            self.y_train = y_train
            self.n_folds = n_folds

        def __len__(self):
            return self.n_folds

        def __iter__(self):
            from wbia_cnn.dataset import stratified_kfold_label_split

            rng = 1809629827
            for _ in stratified_kfold_label_split(
                self.y_train, self.nids_train, n_folds=self.n_folds, rng=rng
            ):
                yield _

    clf_fpath = join(model_dpath, '%s_svc_folds_%s.cPkl' % (target_type, n_folds))
    if not ut.checkpath(clf_fpath):
        """
        Curate strategy:
            Use gridsearch to select a reasonable C=.17
            Then train 10 classifiers with 10 split cross validation.
            This lets us make an "unbias" prediction for each training example.
            Look at predictions for all training examples (predict using only
                classifiers not trained with that point).
            Look at worst worst performing examples.
            Fix any errors that occur.
            Now that the database is better, we learn the actual model.

        Learning strategy:
            * Set aside a set test.
            * The remaining data is the training set.
            * Run Gridsearch with N-fold cross validation on training set to
              look at performance given different hyperparameters of the SVM.
            * Use quadratic interpolation to select a "best" parameter.

            (NOTE grid.predict only uses the best estimator (however it is a refit estimator))

            Train a single SVM using these parameters on all training data.

            Evaluate this SVM on the test set.
        """
        C = None
        if C is None:
            import sklearn
            import sklearn.grid_search
            import sklearn.svm

            # C controls the margin of the hyperplane.
            # Smaller C = Larger Hyperplane
            # So, the larger the C the less willing the SVM will be to get
            # examples wrong.

            param_grid = {
                # 'C': np.linspace(.1, .2, 10),
                'C': [
                    0.0001,
                    0.001,
                    0.005,
                    0.01,
                    0.08,
                    0.1,
                    0.12,
                    0.15,
                    0.17,
                    0.2,
                    0.22,
                    0.5,
                    1.0,
                    100,
                    1000,
                    10000,
                ]
                # 'C': np.linspace(.1, .2, 3),
            }
            clf = sklearn.svm.SVC(
                kernel=str('linear'),
                C=0.17,
                class_weight='balanced',
                decision_function_shape='ovr',
            )
            cv = MyLabelCV(y_train, nids_train, n_folds=n_folds)
            grid = sklearn.grid_search.GridSearchCV(
                clf,
                param_grid=param_grid,
                cv=cv,
                refit=False,
                n_jobs=min(n_folds, 6),
                verbose=10,
            )
            x_train = ds.data.take(train_idx, axis=0)
            y_train = ds.target.take(train_idx, axis=0)
            grid.fit(x_train, y_train)

            for params, mean_score, scores in grid.grid_scores_:
                logger.info(
                    '%0.3f (+/-%0.03f) for %r' % (mean_score, scores.std() * 2, params)
                )

            c_xdata = np.array([t[0]['C'] for t in grid.grid_scores_])
            c_ydata = np.array([t[1] for t in grid.grid_scores_])
            import vtool as vt

            # maxima_x, maxima_y, argmaxima = vt.hist_argmaxima(c_ydata, c_xdata, maxima_thresh=None)
            submaxima_x, submaxima_y = vt.argsubmaxima(c_ydata, c_xdata)
            # pt.draw_hist_subbin_maxima(c_ydata, c_xdata, maxima_thresh=None, remove_endpoints=False)
            C = submaxima_x[0]
            logger.info('C = %r' % (C,))
        else:
            logger.info('C = %r' % (C,))

        clf_all = sklearn.svm.SVC(
            kernel=str('linear'),
            C=C,
            class_weight='balanced',
            decision_function_shape='ovr',
            verbose=10,
        )
        X_train = ds.data.take(train_idx, axis=0)
        clf_all.fit(X_train, y_train)
        ut.save_data(clf_fpath, clf_all.__dict__)
        clf = clf_all
    else:
        clf = sklearn.svm.SVC()
        clf.__dict__.update(**ut.load_data(clf_fpath))

    def classifier_test(clf, X_test, y_test):
        logger.info('[problem] test classifier on %d data points' % (len(test_idx),))
        y_conf = predict_svc_ovr(X_test)
        y_pred = y_conf.argmax(axis=1)
        result = ClfSingleResult(problem.ds, test_idx, y_test, y_pred, y_conf)
        return result

    test_idx = ds._split_idxs['test']
    X_test = ds.data.take(test_idx, axis=0)
    y_test = ds.target.take(test_idx, axis=0)
    result = classifier_test(clf, X_test, y_test)
    result.compile_results()
    result.print_report()

    result_list = [result]

    import pandas as pd

    # import wbia.plottool as pt
    # Combine information from results
    df = pd.concat([r.df for r in result_list])
    df['hardness'] = 1 / df['easiness']
    df['aid'] = ut.take(ds.aids, df.index)
    df['target'] = ut.take(ds.target, df.index)
    df['failed'] = df['pred'] != df['target']

    report = sklearn.metrics.classification_report(
        y_true=df['target'], y_pred=df['pred'], target_names=result.ds.target_names
    )
    logger.info(report)

    confusion = sklearn.metrics.confusion_matrix(df['target'], df['pred'])
    logger.info('Confusion Matrix:')
    logger.info(
        pd.DataFrame(
            confusion,
            columns=[m for m in result.ds.target_names],
            index=['gt ' + m for m in result.ds.target_names],
        )
    )

    # inspect_results(ds, result_list)

    if False:
        if False:
            # train_idx, test_idx = problem.stratified_2sample_idxs()
            train_idx = ds._split_idxs['train']
            test_idx = ds._split_idxs['test']

            # import sklearn.metrics
            # model_dpath = join(ds.dataset_dpath, 'svms')
            # model_fpath = join(model_dpath, target_type + '_svc.cPkl')
            # if ut.checkpath(model_fpath):
            #    clf = sklearn.svm.SVC(kernel=str('linear'), C=.17, class_weight='balanced',
            #                          decision_function_shape='ovr')
            #    clf.__dict__.update(**ut.load_data(model_fpath))
            # else:
            #    clf = problem.fit_new_classifier(train_idx)
            #    ut.ensuredir(model_dpath)
            #    ut.save_data(model_fpath, clf.__dict__)
            result_list = []
            result = problem.test_classifier(clf, test_idx)
            result_list.append(result)

            for result in result_list:
                result.compile_results()

            for result in result_list:
                result.print_report()

            inspect_results(ds, result_list)
        if False:
            result_list = []
            result = problem.test_classifier(clf, train_idx)
            result_list.append(result)
            for result in result_list:
                result.compile_results()
            for result in result_list:
                result.print_report()
            inspect_results(ds, result_list)
        if False:
            result_list = []
            # View support vectors
            support_idxs = clf.support_
            result = problem.test_classifier(clf, support_idxs)
            result_list.append(result)
            for result in result_list:
                result.compile_results()
            for result in result_list:
                result.print_report()
            inspect_results(ds, result_list)


def inspect_results(ds, result_list):
    import pandas as pd
    import wbia.plottool as pt

    pd.set_option('display.max_rows', 20)
    pt.qt4ensure()

    isect_sets = [
        set(s1).intersection(set(s2))
        for s1, s2 in ut.combinations([result.df.index for result in result_list], 2)
    ]
    assert all([len(s) == 0 for s in isect_sets]), 'cv sets should not intersect'

    # Combine information from results
    df = pd.concat([result.df for result in result_list])
    df['hardness'] = 1 / df['easiness']
    df['aid'] = ut.take(ds.aids, df.index)
    df['target'] = ut.take(ds.target, df.index)
    df['failed'] = df['pred'] != df['target']

    report = sklearn.metrics.classification_report(
        y_true=df['target'], y_pred=df['pred'], target_names=result.ds.target_names
    )
    logger.info(report)

    confusion = sklearn.metrics.confusion_matrix(df['target'], df['pred'])
    logger.info('Confusion Matrix:')
    logger.info(
        pd.DataFrame(
            confusion,
            columns=[m for m in result.ds.target_names],
            index=['gt ' + m for m in result.ds.target_names],
        )
    )

    def target_partition(target):
        df_chunk = df if target is None else df[df['target'] == target]
        df_chunk = df_chunk.take(df_chunk['hardness'].argsort())
        return df_chunk

    def grab_subchunk(frac, n, target):
        df_chunk = target_partition(target)
        sl = ut.snapped_slice(len(df_chunk), frac, n)
        logger.info('sl = %r' % (sl,))
        idx = df_chunk.index[sl]
        df_chunk = df_chunk.loc[idx]
        min_frac = sl.start / len(df_chunk)
        max_frac = sl.stop / len(df_chunk)
        min_frac = sl.start
        max_frac = sl.stop
        place_name = 'hardness=%.2f (%d-%d)' % (frac, min_frac, max_frac)
        if target is not None:
            df_chunk.nice = place_name + ' ' + ds.target_names[target]
        else:
            df_chunk.nice = place_name
        return df_chunk

    def grab_subchunk2(df_chunk, frac, n):
        sl = ut.snapped_slice(len(df_chunk), frac, n)
        logger.info('sl = %r' % (sl,))
        idx = df_chunk.index[sl]
        df_chunk = df_chunk.loc[idx]
        min_frac = sl.start / len(df_chunk)
        max_frac = sl.stop / len(df_chunk)
        min_frac = sl.start
        max_frac = sl.stop
        place_name = 'hardness=%.2f (%d-%d)' % (frac, min_frac, max_frac)
        if target is not None:
            df_chunk.nice = place_name + ' ' + ds.target_names[target]
        else:
            df_chunk.nice = place_name
        return df_chunk

    # Look at hardest train cases

    # Look at hardest test cases
    if True:
        # n = 4
        fracs = [0.0, 0.7, 0.8, 0.9, 1.0]
        view_targets = ds.target_labels
        n = 8 // len(view_targets)
    else:
        view_targets = [ut.listfind(ds.target_names.tolist(), 'healthy')]
        # fracs = [0.0, .7, .8, .9, 1.0]
        fracs = [0.45, 0.5, 0.55, 0.6, 0.62]
        fracs = [0.72, 0.82, 0.84, 0.88]
        fracs = [0.73, 0.83, 0.835, 0.89]
        fracs = [0.73, 0.83, 0.835, 0.89]
        fracs = [0.735, 0.833, 0.837, 0.934]
        fracs = [0.2, 0.65, 0.75, 0.85, 0.95]
        fracs = [0.3, 0.4, 0.67, 0.77, 0.87, 0.92]
        n = 8 // len(view_targets)

    if False:
        view_targets = [ut.listfind(ds.target_names.tolist(), 'healthy')]
        target_dfs = [target_partition(target) for target in view_targets]
        critical_points = [np.where(_df['failed'])[0][0] for _df in target_dfs]
        critical_fracs = [_pt / len(_df) for _pt, _df in zip(critical_points, target_dfs)]
        n = 8 * 5
        frac = critical_fracs[0]
        frac += 0.1
        _df = target_dfs[0]
        df_part = grab_subchunk2(_df, frac, n)
        df_chunks = [df_part.iloc[x] for x in ut.ichunks(range(len(df_part)), 8)]
    else:
        df_chunks = [
            grab_subchunk(frac, n, target) for frac in fracs for target in view_targets
        ]

    ibs = ds.ibs
    config = ds.config
    from wbia_cnn import draw_results

    inter = draw_results.make_InteractClasses(
        ibs, config, df_chunks, nCols=len(view_targets)
    )
    inter.start()
